from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import *

import pkg_resources
from collections import defaultdict
from fonduer.snorkel.models import construct_stable_id
from fonduer.snorkel.parser import Parser, ParserConnection
from pathlib import Path

try:
    import spacy
    from spacy.cli import download
    from spacy import util
except:
    raise Exception("spaCy not installed. Use `pip install spacy`.")


class Spacy(Parser):
    '''
    spaCy
    https://spacy.io/

    Models for each target language needs to be downloaded using the
    following command:

    python -m spacy download en

    Default named entity types

    PERSON	    People, including fictional.
    NORP	    Nationalities or religious or political groups.
    FACILITY	Buildings, airports, highways, bridges, etc.
    ORG	        Companies, agencies, institutions, etc.
    GPE	        Countries, cities, states.
    LOC	        Non-GPE locations, mountain ranges, bodies of water.
    PRODUCT	    Objects, vehicles, foods, etc. (Not services.)
    EVENT	    Named hurricanes, battles, wars, sports events, etc.
    WORK_OF_ART	Titles of books, songs, etc.
    LANGUAGE	Any named language.

    DATE	    Absolute or relative dates or periods.
    TIME	    Times smaller than a day.
    PERCENT	    Percentage, including "%".
    MONEY	    Monetary values, including unit.
    QUANTITY	Measurements, as of weight or distance.
    ORDINAL	    "first", "second", etc.
    CARDINAL	Numerals that do not fall under another type.

    '''
    def __init__(self, annotators=['tagger', 'parser', 'entity'],
                 lang='en', num_threads=1, verbose=False):

        super(Spacy, self).__init__(name="spacy")
        self.model = Spacy.load_lang_model(lang)
        self.num_threads = num_threads

        self.pipeline = [proc for _, proc in self.model.__dict__['pipeline']]


    @staticmethod
    def is_package(name):
        """Check if string maps to a package installed via pip.
        name (unicode): Name of package.
        RETURNS (bool): True if installed package, False if not.

        From https://github.com/explosion/spaCy/blob/master/spacy/util.py

        """
        name = name.lower()  # compare package name against lowercase name
        packages = pkg_resources.working_set.by_key.keys()
        for package in packages:
            if package.lower().replace('-', '_') == name:
                return True
        return False


    @staticmethod
    def model_installed(name):
        '''
        Check if spaCy language model is installed

        From https://github.com/explosion/spaCy/blob/master/spacy/util.py

        :param name:
        :return:
        '''
        data_path = util.get_data_path()
        if not data_path or not data_path.exists():
            raise IOError("Can't find spaCy data path: %s" % str(data_path))
        if name in set([d.name for d in data_path.iterdir()]):
            return True
        if Spacy.is_package(name):  # installed as package
            return True
        if Path(name).exists():  # path to model data directory
            return True
        return False


    @staticmethod
    def load_lang_model(lang):
        '''
        Load spaCy language model or download if
        model is available and not installed

        Currenty supported spaCy languages

        en English (50MB)
        de German (645MB)
        fr French (1.33GB)
        es Spanish (377MB)

        :param lang:
        :return:
        '''
        if not Spacy.model_installed(lang):
            download(lang)
        return spacy.load(lang)

    def connect(self):
        return ParserConnection(self)

    def parse(self, document, text):
        '''
        Transform spaCy output to match CoreNLP's default format
        :param document:
        :param text:
        :return:
        '''
        text = self.to_unicode(text)

        doc = self.model.tokenizer(text)
        for proc in self.pipeline:
            proc(doc)
        assert doc.is_parsed

        position = 0
        for sent in doc.sents:
            parts = defaultdict(list)
            text = sent.text

            for i,token in enumerate(sent):
                parts['words'].append(str(token))
                parts['lemmas'].append(token.lemma_)
                parts['pos_tags'].append(token.tag_)
                parts['ner_tags'].append(token.ent_type_ if token.ent_type_ else 'O')
                parts['char_offsets'].append(token.idx)
                parts['abs_char_offsets'].append(token.idx)
                head_idx = 0 if token.head is token else token.head.i - sent[0].i + 1
                parts['dep_parents'].append(head_idx)
                parts['dep_labels'].append(token.dep_)

            # Add null entity array (matching null for CoreNLP)
            parts['entity_cids'] = ['O' for _ in parts['words']]
            parts['entity_types'] = ['O' for _ in parts['words']]

            # make char_offsets relative to start of sentence
            parts['char_offsets'] = [
                p - parts['char_offsets'][0] for p in parts['char_offsets']
            ]
            parts['position'] = position

            # Link the sentence to its parent document object
            parts['document'] = document
            parts['text'] = text

            # Add null entity array (matching null for CoreNLP)
            parts['entity_cids'] = ['O' for _ in parts['words']]
            parts['entity_types'] = ['O' for _ in parts['words']]

            # Assign the stable id as document's stable id plus absolute
            # character offset
            abs_sent_offset = parts['abs_char_offsets'][0]
            abs_sent_offset_end = abs_sent_offset + parts['char_offsets'][-1] + len(parts['words'][-1])
            if document:
                parts['stable_id'] = construct_stable_id(document, 'sentence', abs_sent_offset, abs_sent_offset_end)

            position += 1

            yield parts



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import *
from future.utils import iteritems

from .classifier import Classifier
from numba import jit
import numbskull
from numbskull import NumbSkull
from numbskull.inference import FACTORS
from numbskull.numbskulltypes import Weight, Variable, Factor, FactorToVar
import numpy as np
import random
import scipy.sparse as sparse
from copy import copy
from pandas import DataFrame
from distutils.version import StrictVersion
from six.moves.cPickle import dump, load
import os

DEP_SIMILAR = 0
DEP_FIXING = 1
DEP_REINFORCING = 2
DEP_EXCLUSIVE = 3


class GenerativeModel(Classifier):
    """
    A generative model for data programming for binary classification.

    Supports dependencies among labeling functions.

    :param class_prior: whether to include class label prior factors
    :param lf_prior: whether to include labeling function prior factors
    :param lf_propensity: whether to include labeling function propensity
        factors
    :param lf_class_propensity: whether to include class-specific labeling
        function propensity factors
    :param seed: seed for initializing state of Numbskull variables
    """
    def __init__(self, class_prior=False, lf_prior=False, lf_propensity=False,
        lf_class_propensity=False, seed=271828, name=None):
        self.name = name or self.__class__.__name__
        try:
            numbskull_version = numbskull.__version__
        except:
            numbskull_version = "0.0"
        numbskull_require = "0.1"

        if StrictVersion(numbskull_version) < StrictVersion(numbskull_require):
            raise ValueError(
                "Snorkel requires Numbskull version %s, but version %s is installed." % (numbskull_require, numbskull_version))

        self.class_prior = class_prior
        self.lf_prior = lf_prior
        self.lf_propensity = lf_propensity
        self.lf_class_propensity = lf_class_propensity
        self.weights = None

        self.rng = np.random.RandomState()
        self.rng.seed(seed)
        set_numba_seeds(seed)

    # These names of factor types are for the convenience of several methods
    # that perform the same operations over multiple types, but this class's
    # behavior is not fully specified here. Other methods, such as marginals(),
    # as well as maps defined within methods, require manual adjustments to
    # implement changes.
    #
    # These names are also used by other related classes, such as
    # GenerativeModelParameters
    optional_names = ('lf_prior', 'lf_propensity', 'lf_class_propensity')
    dep_names = (
        'dep_similar', 'dep_fixing', 'dep_reinforcing', 'dep_exclusive'
    )

    def train(self, L, deps=(), LF_acc_prior_weights=None,
        LF_acc_prior_weight_default=1, labels=None, label_prior_weight=5,
        init_deps=0.0, init_class_prior=-1.0, epochs=30, step_size=None, 
        decay=1.0, reg_param=0.1, reg_type=2, verbose=False, truncation=10, 
        burn_in=5, cardinality=None, timer=None, candidate_ranges=None, threads=1):
        """
        Fits the parameters of the model to a data set. By default, learns a
        conditionally independent model. Additional unary dependencies can be
        set to be included in the constructor. Additional pairwise and
        higher-order dependencies can be included as an argument.

        Results are stored as a member named weights, instance of
        snorkel.learning.gen_learning.GenerativeModelWeights.

        :param L: M x N csr_AnnotationMatrix-type label matrix, where there are 
            M candidates labeled by N labeling functions (LFs)
        :param deps: collection of dependencies to include in the model, each 
                     element is a tuple of the form 
                     (LF 1 index, LF 2 index, dependency type),
                     see snorkel.learning.constants
        :param LF_acc_prior_weights: An N-element list of prior weights for the
            LF accuracies (log scale)
        :param LF_acc_prior_weight_default: Default prior for the weight of each 
            LF accuracy; if LF_acc_prior_weights is unset, each LF will have 
            this accuracy prior weight (log scale)
        :param labels: Optional ground truth labels
        :param label_prior_weight: The prior probability that the ground truth 
            labels (if provided) are correct (log scale)
        :param init_deps: initial weight for additional dependencies, except
                          class prior (log scale)
        :param init_class_prior: initial class prior (in log scale), note only
                                 used if class_prior=True in constructor
        :param epochs: number of training epochs
        :param step_size: gradient step size, default is 1 / L.shape[0]
        :param decay: multiplicative decay of step size,
                      step_size_(t+1) = step_size_(t) * decay
        :param reg_param: regularization strength
        :param reg_type: 1 = L1 regularization, 2 = L2 regularization
        :param verbose: whether to write debugging info to stdout
        :param truncation: number of iterations between truncation step for L1
                           regularization
        :param burn_in: number of burn-in samples to take before beginning
                        learning
        :param cardinality: number of possible classes; by default is inferred
            from the label matrix L
        :param timer: stopwatch for profiling, must implement start() and end()
        :param candidate_ranges: Optionally, a list of M sets of integer values,
            representing the possible categorical values that each of the M
            candidates can take. If a label is outside of this range throws an
            error. If None, then each candidate can take any value from 0 to
            cardinality.
        :param threads: the number of threads to use for sampling. Default is 1.
        """
        m, n = L.shape
        step_size = step_size or 0.0001

        # Check to make sure matrix is int-valued
        element_type = type(L[0,0])
        # Note: Other simpler forms of this check often don't work; still not
        # sure why...
        if not issubclass(element_type, np.integer):
            raise ValueError("""Label matrix must have int-type elements, 
                but elements have type %s""" % element_type)

        # Automatically infer cardinality
        # Binary: Values in {-1, 0, 1} [Default]
        # Categorical: Values in {0, 1, ..., K}
        if cardinality is None:
            # If candidate_ranges is provided, use this to determine cardinality
            if candidate_ranges is not None:
                cardinality = max(map(max, candidate_ranges))
            else:
                # This is just an annoying hack for LIL sparse matrices...
                try:
                    lmax = L.max()
                except AttributeError:
                    lmax = L.tocoo().max()

                if lmax > 2:
                    cardinality = lmax
                elif lmax < 2:
                    cardinality = 2
                else:
                    raise ValueError(
                        "L.max() == %s, cannot infer cardinality." % lmax)
            print("Inferred cardinality: %s" % cardinality)
        self.cardinality = cardinality

        # Priors for LFs default to fixed prior value
        # NOTE: Setting default != 0.5 creates a (fixed) factor which increases
        # runtime (by ~0.5x that of a non-fixed factor)...
        if LF_acc_prior_weights is None:
            LF_acc_prior_weights = [LF_acc_prior_weight_default for _ in range(n)]
        else:
            LF_acc_prior_weights = list(copy(LF_acc_prior_weights))

        # LF weights are un-fixed
        is_fixed = [False for _ in range(n)]

        # If supervised labels are provided, add them as a fixed LF with prior
        # Note: For large L this column stack operation could be very
        # inefficient, can consider refactoring...
        if labels is not None:
            labels = labels.reshape(m, 1)
            L = sparse.hstack([L, labels])
            is_fixed.append(True)
            LF_acc_prior_weights.append(label_prior_weight)
            n += 1

        # Reduce overhead of tracking indices by converting L to a CSR sparse matrix.
        L = sparse.csr_matrix(L).copy()

        # If candidate_ranges is provided, remap the values of L using
        # candidate_ranges. This "scoped categorical" approach allows learning
        # and inference to be efficient even with very large cardinality, as
        # we only sample relevant values for each candidate. Also set
        # per-candidate cardinalities according to candidate_ranges if not None,
        # else as constant value.
        self.cardinalities = self.cardinality * np.ones(m, dtype=np.int64)
        self.candidate_ranges = candidate_ranges
        if self.candidate_ranges is not None:
            L, self.cardinalities, _ = self._remap_scoped_categoricals(L, 
                self.candidate_ranges)

        # Shuffle the data points, cardinalities, and candidate_ranges
        idxs = self.rng.permutation(list(range(m)))
        L = L[idxs, :]
        if candidate_ranges is not None:
            self.cardinalities = self.cardinalities[idxs]
            c_ranges_reshuffled = []
            for i in idxs:
                c_ranges_reshuffled.append(self.candidate_ranges[i])
            self.candidate_ranges = c_ranges_reshuffled

        # Compile factor graph
        self._process_dependency_graph(L, deps)
        weight, variable, factor, ftv, domain_mask, n_edges = self._compile(
            L, init_deps, init_class_prior, LF_acc_prior_weights, is_fixed, self.cardinalities)
        fg = NumbSkull(
            n_inference_epoch=0,
            n_learning_epoch=epochs, 
            stepsize=step_size,
            decay=decay,
            reg_param=reg_param,
            regularization=reg_type,
            truncation=truncation,
            quiet=(not verbose),
            verbose=verbose, 
            learn_non_evidence=True,
            burn_in=burn_in,
            nthreads=threads
        )
        fg.loadFactorGraph(weight, variable, factor, ftv, domain_mask, n_edges)

        if timer is not None:
            timer.start()
        fg.learning(out=False)
        if timer is not None:
            timer.end()
        self._process_learned_weights(L, fg, LF_acc_prior_weights, is_fixed)

        # Store info from factor graph
        if self.candidate_ranges is not None:
            self.cardinality_for_stats = int(max(self.cardinalities))
        else:
            self.cardinality_for_stats = self.cardinality
        self.learned_weights = fg.factorGraphs[0].weight_value
        weight, variable, factor, ftv, domain_mask, n_edges =\
            self._compile(sparse.coo_matrix((1, n), L.dtype), init_deps,
                init_class_prior, LF_acc_prior_weights, is_fixed,
                [self.cardinality_for_stats])

        variable["isEvidence"] = False
        weight["isFixed"] = True
        weight["initialValue"] = fg.factorGraphs[0].weight_value

        fg.factorGraphs = []
        fg.loadFactorGraph(weight, variable, factor, ftv, domain_mask, n_edges)

        self.fg = fg
        self.nlf = n
        self.cardinality = cardinality

    def _remap_scoped_categoricals(self, L_in, candidate_ranges):
        """
        Remap the values of each individual candidate so that they have dense
        support, returning the remapped label matrix, cardinalities, and
        inverse mapping.
        """
        L = L_in.copy()
        m, n = L.shape
        cardinalities = np.ones(m)
        mappings = []
        for i in range(m):
            c_range = candidate_ranges[i]

            # Confirm that the candidate range has only unique values
            assert len(c_range) == len(set(c_range))
            cardinalities[i] = len(c_range)

            # Create the inverse mapping
            mappings.append(dict([(a + 1, b) for a, b in enumerate(c_range)]))

            # Re-map the values of L[i, :]
            # Assumes L is csr_sparse format at this point
            for j in range(L[i].data.shape[0]):
                val = L[i].data[j]
                if val not in c_range:
                    raise ValueError("""Value {0} is not in supplied range 
                        for candidate at index {1}""".format(val, i))
                L[i, L[i].indices[j]] = c_range.index(val) + 1
        return L, cardinalities, mappings

    def learned_lf_stats(self):
        """
        Provides a summary of what the model has learned about the labeling
        functions. For each labeling function, estimates of the following
        are provided:

            Abstain
            Accuracy
            Coverage

            [Following are only available for binary tasks]
            True  Positive (TP)
            False Positive (FP)
            True  Negative (TN)
            False Negative (FN)

        For scoped categoricals, the information provided is for the maximum
        observed cardinality of any single data point.

        WARNING: This uses Gibbs sampling to estimate these values. This will
                 tend to mix poorly when there are many very accurate labeling
                 functions. In this case, this function will assume that the
                 classes are approximately balanced.
        """
        if self.fg is None:
            raise ValueError(
                "Must fit model with train() before computing diagnostics.")

        burnin = 500
        trials = 5000
        cardinality = self.cardinality_for_stats
        count = np.zeros((self.nlf, cardinality, cardinality + 1))

        for true_label in range(cardinality):
            for i in range(self.nlf + 1):
                self.fg.factorGraphs[0].var_value[0, i] = true_label
            self.fg.factorGraphs[0].inference(burnin, 0, True)
            for i in range(trials):
                self.fg.factorGraphs[0].inference(0, 1, True)
                y = self.fg.factorGraphs[0].var_value[0, 0]
                for j in range(self.nlf):
                    lf = self.fg.factorGraphs[0].var_value[0, j + 1]
                    count[j, y, lf] += 1

        count /= cardinality * trials

        # Compute summary stats to return to user
        stats = []
        for i in range(self.nlf):
            if cardinality == 2:
                tp = count[i, 1, 1]
                fp = count[i, 0, 1]
                tn = count[i, 0, 0]
                fn = count[i, 1, 0]
                coverage = 1 - (count[i, 0, 2] + count[i, 1, 2])
                stats.append({
                    "Precision": tp / (tp + fp),
                    "Recall": tp / count[i, 1, :].sum(),
                    "Accuracy": (tp + tn) / coverage,
                    "Coverage": coverage
                    })
            else:
                correct = sum([count[i, j, j] for j in range(cardinality)])
                coverage = 1 - sum([count[i, j, cardinality]
                    for j in range(cardinality)])
                stats.append({
                    "Accuracy": correct / coverage,
                    "Coverage": coverage
                })

        return DataFrame(stats)

    def marginals(self, L, candidate_ranges=None, batch_size=None):
        """
        Given an M x N label matrix, returns marginal probabilities for each
        candidate, depending on classification setting:

            - Binary: Returns M-dim array representing the marginal probability
                of each candidate being True

            - Categorical (cardinality = K): Returns M x K dense matrix
                representing the marginal probabilities of each candidate being
                each class.

            - Scoped Categorical (cardinality = K, cardinality_ranges not None):
                Returns an M x K *sparse* matrix of marginals.

        In the categorical setting, the K values (columns in the marginals
        matrix) correspond to indices of the Candidate values defined.
        """
        m, n = L.shape
        if self.weights is None:
            raise ValueError("""Must fit model with train() before computing 
                marginal probabilities.""")

        # Binary classification setting
        if self.cardinality == 2:
            marginals = np.ndarray(L.shape[0], dtype=np.float64)

            for i in range(L.shape[0]):
                logp_true = self.weights.class_prior
                logp_false = -1 * self.weights.class_prior

                l_i = L[i].tocoo()

                for l_index1 in range(l_i.nnz):
                    data_j, j = l_i.data[l_index1], l_i.col[l_index1]
                    if data_j == 1:
                        logp_true  += self.weights.lf_accuracy[j]
                        logp_false -= self.weights.lf_accuracy[j]
                        logp_true  += self.weights.lf_class_propensity[j]
                        logp_false -= self.weights.lf_class_propensity[j]
                    elif data_j == -1:
                        logp_true  -= self.weights.lf_accuracy[j]
                        logp_false += self.weights.lf_accuracy[j]
                        logp_true  += self.weights.lf_class_propensity[j]
                        logp_false -= self.weights.lf_class_propensity[j]
                    else:
                        ValueError("""Illegal value at %d, %d: %d.
                            Must be in {-1, 0, 1}.""" % (i, j, data_j))

                    for l_index2 in range(l_i.nnz):
                        data_k, k = l_i.data[l_index2], l_i.col[l_index2]
                        if j != k:
                            if data_j == -1 and data_k == 1:
                                logp_true += self.weights.dep_fixing[j, k]
                            elif data_j == 1 and data_k == -1:
                                logp_false += self.weights.dep_fixing[j, k]

                            if data_j == 1 and data_k == 1:
                                logp_true += self.weights.dep_reinforcing[j, k]
                            elif data_j == -1 and data_k == -1:
                                logp_false += self.weights.dep_reinforcing[j, k]

                marginals[i] = 1 / (1 + np.exp(logp_false - logp_true))
            return marginals

        # Categorical setting
        else:
            all_marginals = []

            # Handle the scoped categorical case, otherwise get cardinalities
            # from self.cardinality
            if candidate_ranges is not None:
                L, cardinalities, mappings = self._remap_scoped_categoricals(L, 
                    candidate_ranges)
            else:
                cardinalities = self.cardinality * np.ones(m)

            # Get the marginal (posterior) probability for each candidate
            for i in range(m):
                cardinality = int(cardinalities[i])
                marginals = np.zeros(cardinality, dtype=np.float64)
                # NB: class priors not currently available for categoricals
                l_i = L[i].tocoo()
                for l_index1 in range(l_i.nnz):
                    data_j, j = l_i.data[l_index1], l_i.col[l_index1]
                    if (data_j != 0):
                        if not 1 <= data_j <= cardinality:
                            raise ValueError(
                                """Illegal value at %d, %d: %d. Must be in 0 to 
                                %d.""" % (i, j, data_j, cardinality))
                        # NB: LF class propensity not currently available
                        # for categoricals
                        marginals[int(data_j - 1)] += \
                            2 * self.weights.lf_accuracy[j]
                            
                # NB: fixing and reinforcing not available for categoricals
                # Get softmax
                exps = np.exp(marginals)
                marginals = exps / exps.sum()
                all_marginals.append(marginals)

            # If candidate_ranges not None, remap back to original values and
            # return as sparse matrix
            if candidate_ranges is not None:
                M = sparse.coo_matrix((m, self.cardinality), dtype=np.float64)
                for i, marginals in enumerate(all_marginals):
                    for j, p in enumerate(marginals):
                        M[i, mappings[i][j]] = p
            else:
                M = np.vstack(all_marginals)
            return M

    def _process_dependency_graph(self, L, deps):
        """
        Processes an iterable of triples that specify labeling function dependencies.

        The first two elements of the triple are the labeling functions to be modeled as dependent. The labeling
        functions are specified using their column indices in `L`. The third element is the type of dependency.
        Options are :const:`DEP_SIMILAR`, :const:`DEP_FIXING`, :const:`DEP_REINFORCING`, and :const:`DEP_EXCLUSIVE`.

        The results are :class:`scipy.sparse.csr_matrix` objects that represent directed adjacency matrices. They are
        set as various GenerativeModel members, two for each type of dependency, e.g., `dep_similar` and `dep_similar_T`
        (its transpose for efficient inverse lookups).

        :param deps: iterable of tuples of the form (lf_1, lf_2, type)
        """
        dep_name_map = {
            DEP_SIMILAR: 'dep_similar',
            DEP_FIXING: 'dep_fixing',
            DEP_REINFORCING: 'dep_reinforcing',
            DEP_EXCLUSIVE: 'dep_exclusive'
        }

        for dep_name in GenerativeModel.dep_names:
            setattr(self, dep_name, sparse.lil_matrix((L.shape[1], L.shape[1])))

        for lf1, lf2, dep_type in deps:
            if lf1 == lf2:
                raise ValueError("Invalid dependency. Labeling function cannot depend on itself.")

            if dep_type in dep_name_map:
                dep_mat = getattr(self, dep_name_map[dep_type])
            else:
                raise ValueError("Unrecognized dependency type: " + str(dep_type))

            dep_mat[lf1, lf2] = 1

        for dep_name in GenerativeModel.dep_names:
            setattr(self, dep_name, getattr(self, dep_name).tocoo(copy=True))

    def _compile(self, L, init_deps, init_class_prior, LF_acc_prior_weights, is_fixed, cardinalities):
        """Compiles a generative model based on L and the current labeling function
        dependencies.
        """
        m, n = L.shape

        n_weights = 1 if self.class_prior else 0

        self.hasPrior = [i != 0 for i in LF_acc_prior_weights]
        nPrior = sum(self.hasPrior)
        nUnFixed = sum([not i for i in is_fixed])

        n_weights += nPrior
        n_weights += nUnFixed
        for optional_name in GenerativeModel.optional_names:
            if getattr(self, optional_name):
                n_weights += n
        for dep_name in GenerativeModel.dep_names:
            n_weights += getattr(self, dep_name).getnnz()

        n_vars = m * (n + 1)
        n_factors = m * n_weights

        n_edges = 1 if self.class_prior else 0
        n_edges += 2 * (nPrior + nUnFixed)
        if self.lf_prior:
            n_edges += n
        if self.lf_propensity:
            n_edges += n
        if self.lf_class_propensity:
            n_edges += 2 * n
        n_edges += 2 * self.dep_similar.getnnz() + 3 * self.dep_fixing.getnnz() + \
                   3 * self.dep_reinforcing.getnnz() + 2 * self.dep_exclusive.getnnz()
        n_edges *= m

        weight = np.zeros(n_weights, Weight)
        variable = np.zeros(n_vars, Variable)
        factor = np.zeros(n_factors, Factor)
        ftv = np.zeros(n_edges, FactorToVar)
        domain_mask = np.zeros(n_vars, np.bool)

        #
        # Compiles weight matrix
        #
        if self.class_prior:
            weight[0]['isFixed'] = False
            weight[0]['initialValue'] = np.float64(init_class_prior)
            w_off = 1
        else:
            w_off = 0

        for i in range(n):
            # Prior on LF acc
            if self.hasPrior[i]:
                weight[w_off]['isFixed'] = True
                weight[w_off]['initialValue'] = LF_acc_prior_weights[i]
                w_off += 1
            # Learnable acc for LF
            if (not is_fixed[i]):
                weight[w_off]['isFixed'] = False

                # Note: Because we're not doing exact gradient descent, don't
                # need to add any random noise to initial values here
                # Setting to 0 = setting to prior value
                weight[w_off]['initialValue'] = np.float64(0)
                w_off += 1

        for i in range(w_off, weight.shape[0]):
            weight[i]['isFixed'] = False
            weight[i]['initialValue'] = np.float64(init_deps)

        #
        # Compiles variable matrix
        #
        # Internal representation:
        #   True Class:         0 to (cardinality - 1) are the classes
        #   Labeling functions: 0 to (cardinality - 1) are the classes
        #                       cardinality is abstain
        # Candidates (variables)
        for i in range(m):
            variable[i]['isEvidence'] = False
            variable[i]['initialValue'] = self.rng.randint(cardinalities[i])
            variable[i]["dataType"] = 0
            variable[i]["cardinality"] = cardinalities[i]

        # LF label variables -- initial loop to set all variables
        for i in range(m):
            for j in range(n):
                index = m + n * i + j
                variable[index]["isEvidence"] = 1
                variable[index]["dataType"] = 0
                variable[index]["cardinality"] = cardinalities[i] + 1
                
                # Default to abstain
                variable[index]["initialValue"] = cardinalities[i]

        # LF labels -- now set the non-zero labels
        L_coo = L.tocoo()
        for L_index in range(L_coo.nnz):
            data, i, j = L_coo.data[L_index], L_coo.row[L_index], L_coo.col[L_index]
            index = m + n * i + j

            # Note: Here we need to use the overall cardinality to handle, since
            # with candidate_ranges not None and self.cardinality > 2, some
            # candidates could have cardinality == 2...
            if (self.cardinality == 2):
                if data == 1:
                    variable[index]["initialValue"] = 1
                elif data == 0:
                    variable[index]["initialValue"] = 2
                elif data == -1:
                    variable[index]["initialValue"] = 0
                else:
                    raise ValueError("Invalid labeling function output in cell (%d, %d): %d. "
                                     "Valid values are 1, 0, and -1. " % (i, j, data))
            else:
                if data == 0:
                    variable[index]["initialValue"] = cardinalities[i]
                elif 1 <= data <= cardinalities[i]:
                    variable[index]["initialValue"] = data - 1
                else:
                    raise ValueError("Invalid labeling function output in cell (%d, %d): %d. "
                                     "Valid values are 0 to %d. " % (i, j, data, self.cardinalities[i]))

        #
        # Compiles factor and ftv matrices
        #
        # Class prior
        if self.class_prior:
            if self.cardinality != 2:
                raise NotImplementedError("Class Prior not implemented for categorical classes.")
            for i in range(m):
                factor[i]["factorFunction"] = FACTORS["DP_GEN_CLASS_PRIOR"]
                factor[i]["weightId"] = 0
                factor[i]["featureValue"] = 1
                factor[i]["arity"] = 1
                factor[i]["ftv_offset"] = i

                ftv[i]["vid"] = i

            f_off = m
            ftv_off = m
            w_off = 1
        else:
            f_off = 0
            ftv_off = 0
            w_off = 0

        # Factors over labeling function outputs
        nfactors_for_lf = [(int(self.hasPrior[i]) + int(not is_fixed[i])) for i in range(n)]
        f_off, ftv_off, w_off = self._compile_output_factors(L, factor, f_off, ftv, ftv_off, w_off, "DP_GEN_LF_ACCURACY",
                                                             (lambda m, n, i, j: i, lambda m, n, i, j: m + n * i + j), nfactors_for_lf)

        optional_name_map = {
            'lf_prior':
                ('DP_GEN_LF_PRIOR', (
                    lambda m, n, i, j: m + n * i + j,)),
            'lf_propensity':
                ('DP_GEN_LF_PROPENSITY', (
                    lambda m, n, i, j: m + n * i + j,)),
            'lf_class_propensity':
                ('DP_GEN_LF_CLASS_PROPENSITY', (
                    lambda m, n, i, j: i,
                    lambda m, n, i, j: m + n * i + j)),
        }

        for optional_name in GenerativeModel.optional_names:
            if getattr(self, optional_name):
                if optional_name != 'lf_propensity' and self.cardinality != 2:
                    raise NotImplementedError(optional_name + " not implemented for categorical classes.")
                f_off, ftv_off, w_off = self._compile_output_factors(L, factor, f_off, ftv, ftv_off, w_off,
                                                                     optional_name_map[optional_name][0],
                                                                     optional_name_map[optional_name][1])

        # Factors for labeling function dependencies
        dep_name_map = {
            'dep_similar':
                ('DP_GEN_DEP_SIMILAR', (
                    lambda m, n, i, j, k: m + n * i + j,
                    lambda m, n, i, j, k: m + n * i + k)),
            'dep_fixing':
                ('DP_GEN_DEP_FIXING', (
                    lambda m, n, i, j, k: i,
                    lambda m, n, i, j, k: m + n * i + j,
                    lambda m, n, i, j, k: m + n * i + k)),
            'dep_reinforcing':
                ('DP_GEN_DEP_REINFORCING', (
                    lambda m, n, i, j, k: i,
                    lambda m, n, i, j, k: m + n * i + j,
                    lambda m, n, i, j, k: m + n * i + k)),
            'dep_exclusive':
                ('DP_GEN_DEP_EXCLUSIVE', (
                    lambda m, n, i, j, k: m + n * i + j,
                    lambda m, n, i, j, k: m + n * i + k))
        }

        CATEGORICAL_DEPS = ['dep_similar', 'dep_exclusive']
        for dep_name in GenerativeModel.dep_names:
            mat = getattr(self, dep_name)
            if mat.nnz > 0:
                if dep_name not in CATEGORICAL_DEPS and self.cardinality != 2:
                    raise NotImplementedError(
                        dep_name + " not implemented for categorical classes.")
                for i in range(len(mat.data)):
                    f_off, ftv_off, w_off = self._compile_dep_factors(L, factor, 
                        f_off, ftv, ftv_off, w_off, mat.row[i], mat.col[i],
                        dep_name_map[dep_name][0], dep_name_map[dep_name][1])

        return weight, variable, factor, ftv, domain_mask, n_edges

    def _compile_output_factors(self, L, factors, factors_offset, ftv, 
        ftv_offset, weight_offset, factor_name, vid_funcs,
        nfactors_for_lf=None):
        """
        Compiles factors over the outputs of labeling functions, i.e., for which
        there is one weight per labeling function and one factor per labeling 
        function-candidate pair.
        """
        m, n = L.shape

        if nfactors_for_lf == None:
            nfactors_for_lf = [1 for i in range(n)]

        factors_index = factors_offset
        ftv_index = ftv_offset
        for i in range(m):
            w_off = weight_offset
            for j in range(n):
                for k in range(nfactors_for_lf[j]):
                    factors[factors_index]["factorFunction"] = FACTORS[factor_name]
                    factors[factors_index]["weightId"] = w_off
                    factors[factors_index]["featureValue"] = 1
                    factors[factors_index]["arity"] = len(vid_funcs)
                    factors[factors_index]["ftv_offset"] = ftv_index

                    factors_index += 1
                    w_off += 1

                    for vid_func in vid_funcs:
                        ftv[ftv_index]["vid"] = vid_func(m, n, i, j)
                        ftv_index += 1

        return factors_index, ftv_index, w_off

    def _compile_dep_factors(self, L, factors, factors_offset, ftv, ftv_offset, weight_offset, j, k, factor_name, vid_funcs):
        """
        Compiles factors for dependencies between pairs of labeling functions (possibly also depending on the latent
        class label).
        """
        m, n = L.shape

        for i in range(m):
            factors_index = factors_offset + i
            ftv_index = ftv_offset + len(vid_funcs) * i

            factors[factors_index]["factorFunction"] = FACTORS[factor_name]
            factors[factors_index]["weightId"] = weight_offset
            factors[factors_index]["featureValue"] = 1
            factors[factors_index]["arity"] = len(vid_funcs)
            factors[factors_index]["ftv_offset"] = ftv_index

            for i_var, vid_func in enumerate(vid_funcs):
                ftv[ftv_index + i_var]["vid"] = vid_func(m, n, i, j, k)

        return factors_offset + m, ftv_offset + len(vid_funcs) * m, weight_offset + 1

    def _process_learned_weights(self, L, fg, LF_acc_prior_weights, is_fixed):
        _, n = L.shape

        w = fg.getFactorGraph().getWeights()
        weights = GenerativeModelWeights(n)

        if self.class_prior:
            weights.class_prior = w[0]
            w_off = 1
        else:
            w_off = 0

        weights.lf_accuracy = np.zeros((n,))
        for i in range(n):
            # Prior on LF acc
            if self.hasPrior[i]:
                weights.lf_accuracy[i] += w[w_off]
                w_off += 1
            # Learnable acc for LF
            if (not is_fixed[i]):
                weights.lf_accuracy[i] += w[w_off]
                w_off += 1

        for optional_name in GenerativeModel.optional_names:
            if getattr(self, optional_name):
                setattr(weights, optional_name, np.copy(w[w_off:w_off + n]))
                w_off += n

        for dep_name in self.dep_names:
            mat = getattr(self, dep_name)
            weight_mat = sparse.lil_matrix((n, n))

            for i in range(len(mat.data)):
                if w[w_off] != 0:
                    weight_mat[mat.row[i], mat.col[i]] = w[w_off]
                w_off += 1

            setattr(weights, dep_name, weight_mat.tocsr(copy=True))

        self.weights = weights

    def save(self, model_name=None, save_dir='checkpoints', verbose=True):
        """Save current model."""
        model_name = model_name or self.name
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
        
        # Save generative model weights
        save_path = os.path.join(save_dir, "{0}.weights.pkl".format(model_name))
        with open(save_path, 'wb') as f:
            dump(self.weights, f)

        # Save other model hyperparameters needed to rebuild model
        save_path2 = os.path.join(save_dir, "{0}.hps.pkl".format(model_name))
        with open(save_path2, 'wb') as f:
            dump({
                'cardinality': self.cardinality,
                'cardinality_for_stats': self.cardinality_for_stats
            }, f)

        if verbose:
            print("[{0}] Model saved as <{1}>.".format(self.name, model_name))

    def load(self, model_name=None, save_dir='checkpoints', verbose=True):
        """Load model."""
        model_name = model_name or self.name
        save_path = os.path.join(save_dir, "{0}.weights.pkl".format(model_name))
        with open(save_path, 'rb') as f:
            self.weights = load(f)
        save_path2 = os.path.join(save_dir, "{0}.hps.pkl".format(model_name))
        with open(save_path2, 'rb') as f:
            hps = load(f)
            for k, v in iteritems(hps):
                setattr(self, k, v)
        if verbose:
            print("[{0}] Model <{1}> loaded.".format(self.name, model_name))


class GenerativeModelWeights(object):

    def __init__(self, n):
        self.n = n
        self.class_prior = 0.0
        self.lf_accuracy = np.zeros(n, dtype=np.float64)
        for optional_name in GenerativeModel.optional_names:
            setattr(self, optional_name, np.zeros(n, dtype=np.float64))

        for dep_name in GenerativeModel.dep_names:
            setattr(self, dep_name, sparse.lil_matrix((n, n), dtype=np.float64))

    def is_sign_sparsistent(self, other, threshold=0.1):
        if self.n != other.n:
            raise ValueError("Dimension mismatch. %d versus %d" % (self.n, other.n))

        if not self._weight_is_sign_sparsitent(self.class_prior, other.class_prior, threshold):
            return False

        for i in range(self.n):
            if not self._weight_is_sign_sparsitent(
                    self.lf_accuracy[i], other.lf_accuracy[i], threshold):
                return False

        for name in GenerativeModel.optional_names:
            for i in range(self.n):
                if not self._weight_is_sign_sparsitent(
                        getattr(self, name)[i], getattr(other, name)[i], threshold):
                    return False

        for name in GenerativeModel.dep_names:
            for i in range(self.n):
                for j in range(self.n):
                    if not self._weight_is_sign_sparsitent(
                            getattr(self, name)[i, j], getattr(other, name)[i, j], threshold):
                        return False

        return True

    def _weight_is_sign_sparsitent(self, w1, w2, threshold):
        if abs(w1) <= threshold and abs(w2) <= threshold:
            return True
        elif w1 > threshold and w2 > threshold:
            return True
        elif w1 < -1 * threshold and w2 < -1 * threshold:
            return True
        else:
            return False


@jit
def set_numba_seeds(seed):
    np.random.seed(seed)
    random.seed(seed)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import *

def create_serialized_candidate_view(session, C, verbose=False):
    """Creates a view in the database for a Candidate sub-class C defined over
    Span contexts, which are direct children of a single sentence.

    Creates VIEW with schema:
        candidate.id, candidate.split, span0.*, ..., spanK.*, sentence.*

    NOTE: This limited functionality should be expanded for arbitrary context
    trees. Also this should be made more dialect-independent.
    """
    selects, froms, joins = [], [], []
    for i, arg in enumerate(C.__argnames__):
        # Select the context CID
        selects.append("{0}.{1}_cid".format(C.__tablename__, arg))
        # Then all of its span columns
        selects.append("span{0}.*".format(i))
        froms.append("span AS span{0}".format(i))
        joins.append("{0}.{1}_id = span{2}.id".format(C.__tablename__, arg, i))

    sql = """
    CREATE VIEW IF NOT EXISTS {0}_serialized AS
        SELECT
            candidate.id,
            candidate.split,
            {1},
            sentence.*
        FROM
            candidate,
            {0},
            {2},
            sentence
        WHERE
            candidate.id = {0}.id
            AND sentence.id = span0.sentence_id
            AND {3}
    """.format(
        C.__tablename__,
        ", ".join(selects),
        ", ".join(froms),
        " AND ".join(joins)
    )
    if verbose:
        print("Creating view...")
        print(sql)
    session.execute(sql)


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import *
from future.utils import iteritems

import numpy as np
from pandas import DataFrame, Series
import scipy.sparse as sparse
from sqlalchemy.sql import bindparam, select

from .models import (
    GoldLabel, GoldLabelKey, Label, LabelKey, Feature, FeatureKey, Candidate,
    Marginal
)
from .models.meta import new_sessionmaker
from .udf import UDF, UDFRunner
from .utils import (
    matrix_conflicts,
    matrix_coverage,
    matrix_overlaps,
    matrix_tp,
    matrix_fp,
    matrix_fn,
    matrix_tn
)


class csr_AnnotationMatrix(sparse.csr_matrix):
    """
    An extension of the scipy.sparse.csr_matrix class for holding sparse annotation matrices
    and related helper methods.
    """
    def __init__(self, arg1, **kwargs):
        # Note: Currently these need to return None if unset, otherwise matrix copy operations break...
        self.candidate_index    = kwargs.pop('candidate_index', None)
        self.row_index          = kwargs.pop('row_index', None)
        self.annotation_key_cls = kwargs.pop('annotation_key_cls', None)
        self.key_index          = kwargs.pop('key_index', None)
        self.col_index          = kwargs.pop('col_index', None)

        # Note that scipy relies on the first three letters of the class to define matrix type...
        super(csr_AnnotationMatrix, self).__init__(arg1, **kwargs)

    def get_candidate(self, session, i):
        """Return the Candidate object corresponding to row i"""
        return session.query(Candidate).filter(Candidate.id == self.row_index[i]).one()

    def get_row_index(self, candidate):
        """Return the row index of the Candidate"""
        return self.candidate_index[candidate.id]

    def get_key(self, session, j):
        """Return the AnnotationKey object corresponding to column j"""
        return session.query(self.annotation_key_cls)\
                .filter(self.annotation_key_cls.id == self.col_index[j]).one()

    def get_col_index(self, key):
        """Return the cow index of the AnnotationKey"""
        return self.key_index[key.id]

    def _get_sliced_indexes(self, s, axis, index, inv_index):
        """
        Remaps the indexes between matrix rows/cols and candidates/keys.
        Note: This becomes a massive performance bottleneck if not implemented
        properly, so be careful of changing!
        """
        if isinstance(s, slice):
            # Check for empty slice
            if s.start is None and s.stop is None:
                return index, inv_index
            else:
                idxs = np.arange(self.shape[axis])[s]
        elif isinstance(s, int):
            idxs = np.array([s])
        else: # s is an array of ints
            idxs = s
            # If s is the entire slice, skip the remapping step
            if np.array_equal(idxs, list(range(len(idxs)))):
                return index, inv_index

        index_new, inv_index_new = {}, {}
        for i_new, i in enumerate(idxs):
            k = index[i]
            index_new[i_new] = k
            inv_index_new[k] = i_new
        return index_new, inv_index_new

    def __getitem__(self, key):
        X = super(csr_AnnotationMatrix, self).__getitem__(key)

        # If X is an integer or float value, just return it
        if type(X) in [int, float] or issubclass(type(X), np.integer)\
            or issubclass(type(X), np.float):
            return X
        # If X is a matrix, make sure it stays a csr_AnnotationMatrix
        elif not isinstance(X, csr_AnnotationMatrix):
            X = csr_AnnotationMatrix(X)
        # X must be a matrix, so update appropriate csr_AnnotationMatrix fields
        X.annotation_key_cls = self.annotation_key_cls
        row_slice, col_slice = self._unpack_index(key)
        X.row_index, X.candidate_index = self._get_sliced_indexes(
            row_slice, 0, self.row_index, self.candidate_index)
        X.col_index, X.key_index = self._get_sliced_indexes(
            col_slice, 1, self.col_index, self.key_index)
        return X

    def stats(self):
        """Return summary stats about the annotations"""
        raise NotImplementedError()


try:
    class csr_LabelMatrix(csr_AnnotationMatrix):

        def lf_stats(self, session, labels=None, est_accs=None):
            """Returns a pandas DataFrame with the LFs and various per-LF statistics"""
            lf_names = [self.get_key(session, j).name for j in range(self.shape[1])]

            # Default LF stats
            col_names = ['j', 'Coverage', 'Overlaps', 'Conflicts']
            d = {
                'j'         : list(range(self.shape[1])),
                'Coverage'  : Series(data=matrix_coverage(self), index=lf_names),
                'Overlaps'  : Series(data=matrix_overlaps(self), index=lf_names),
                'Conflicts' : Series(data=matrix_conflicts(self), index=lf_names)
            }
            if labels is not None:
                col_names.extend(['TP', 'FP', 'FN', 'TN', 'Empirical Acc.'])
                ls = np.ravel(labels.todense() if sparse.issparse(labels) else labels)
                tp = matrix_tp(self, ls)
                fp = matrix_fp(self, ls)
                fn = matrix_fn(self, ls)
                tn = matrix_tn(self, ls)
                ac = (tp+tn) / (tp+tn+fp+fn)
                d['Empirical Acc.'] = Series(data=ac, index=lf_names)
                d['TP']             = Series(data=tp, index=lf_names)
                d['FP']             = Series(data=fp, index=lf_names)
                d['FN']             = Series(data=fn, index=lf_names)
                d['TN']             = Series(data=tn, index=lf_names)

            if est_accs is not None:
                col_names.append('Learned Acc.')
                d['Learned Acc.'] = est_accs
                d['Learned Acc.'].index = lf_names
            return DataFrame(data=d, index=lf_names)[col_names]

# This is a hack for getting the documentation to build...
except:
    class csr_LabelMatrix(object):
        def lf_stats(self, session, labels=None, est_accs=None):
            return None


class Annotator(UDFRunner):
    """Abstract class for annotating candidates and persisting these annotations to DB"""
    def __init__(self, annotation_class, annotation_key_class, f_gen):
        self.annotation_class     = annotation_class
        self.annotation_key_class = annotation_key_class
        super(Annotator, self).__init__(AnnotatorUDF,
                                        annotation_class=annotation_class,
                                        annotation_key_class=annotation_key_class,
                                        f_gen=f_gen)

    def apply(self, split=0, key_group=0, replace_key_set=True, cids_query=None,
        **kwargs):
        # If we are replacing the key set, make sure the reducer key id cache is cleared!
        if replace_key_set:
            self.reducer.key_cache = {}

        # Get the cids based on the split, and also the count
        SnorkelSession = new_sessionmaker()
        session = SnorkelSession()
        cids_query = cids_query or session.query(Candidate.id)\
                                          .filter(Candidate.split == split)

        # Note: In the current UDFRunner implementation, we load all these into memory and fill a
        # multiprocessing JoinableQueue with them before starting... so might as well load them here and pass in.
        # Also, if we try to pass in a query iterator instead, with AUTOCOMMIT on, we get a TXN error...
        cids       = cids_query.all()
        cids_count = len(cids)

        # Run the Annotator
        super(Annotator, self).apply(cids, split=split, key_group=key_group,
            replace_key_set=replace_key_set, cids_query=cids_query,
            count=cids_count, **kwargs)

        # Load the matrix
        return self.load_matrix(session, split=split, cids_query=cids_query,
            key_group=key_group)

    def clear(self, session, split=0, key_group=0, replace_key_set=True,
        cids_query=None, **kwargs):
        """
        Deletes the Annotations for the Candidates in the given split.
        If replace_key_set=True, deletes *all* Annotations (of this Annotation sub-class)
        and also deletes all AnnotationKeys (of this sub-class)
        """
        query = session.query(self.annotation_class)

        # If replace_key_set=False, then we just delete the annotations for candidates in our split
        if not replace_key_set:
            sub_query = cids_query or session.query(Candidate.id)\
                                             .filter(Candidate.split == split)
            sub_query = sub_query.subquery()
            query = query.filter(self.annotation_class.candidate_id.in_(sub_query))
        query.delete(synchronize_session='fetch')

        # If we are creating a new key set, delete all old annotation keys
        if replace_key_set:
            query = session.query(self.annotation_key_class)
            query = query.filter(self.annotation_key_class.group == key_group)
            query.delete(synchronize_session='fetch')

    def apply_existing(self, split=0, key_group=0, cids_query=None, **kwargs):
        """Alias for apply that emphasizes we are using an existing AnnotatorKey set."""
        return self.apply(split=split, key_group=key_group,
            replace_key_set=False, cids_query=cids_query, **kwargs)

    def load_matrix(self, session, split=0, key_group=0, cids_query=None,
        **kwargs):
        raise NotImplementedError()


class AnnotatorUDF(UDF):
    def __init__(self, annotation_class, annotation_key_class, f_gen, **kwargs):
        self.annotation_class     = annotation_class
        self.annotation_key_class = annotation_key_class

        # AnnotatorUDF relies on a *generator function* which yields annotations
        # given a candidate input
        # NB: inspect.isgeneratorfunction is not sufficient to check if f_ger
        # is a generator (does not work with fns that wrap gen, e.g. partial)
        # So no check here at the moment...
        self.anno_generator = f_gen

        # For caching key ids during the reduce step
        self.key_cache = {}

        super(AnnotatorUDF, self).__init__(**kwargs)

    def apply(self, cid, **kwargs):
        """
        Applies a given function to a Candidate, yielding a set of Annotations as key_name, value pairs

        Note: Accepts a candidate _id_ as argument, because of issues with putting Candidate subclasses
        into Queues (can't pickle...)
        """
        seen = set()
        cid = cid[0]
        c    = self.session.query(Candidate).filter(Candidate.id == cid).one()
        for key_name, value in self.anno_generator(c):

            # Note: Make sure no duplicates emitted here!
            if (cid, key_name) not in seen:
                seen.add((cid, key_name))
                yield cid, key_name, value

    def reduce(self, y, clear, key_group, replace_key_set, **kwargs):
        """
        Inserts Annotations into the database.
        For Annotations with unseen AnnotationKeys (in key_group, if not None), either adds these
        AnnotationKeys if create_new_keyset is True, else skips these Annotations.
        """
        cid, key_name, value = y

        # Prepares queries
        # Annoation updating only needs to be done if clear=False
        if not clear:
            anno_update_query = self.annotation_class.__table__.update()
            anno_update_query = anno_update_query.where(self.annotation_class.candidate_id == bindparam('cid'))
            anno_update_query = anno_update_query.where(self.annotation_class.key_id == bindparam('kid'))
            anno_update_query = anno_update_query.values(value=bindparam('value'))

        # We only need to insert AnnotationKeys if replace_key_set=True
        # Note that in current configuration, we never update AnnotationKeys!
        if replace_key_set:
            key_insert_query = self.annotation_key_class.__table__.insert()

        # If we are replacing the AnnotationKeys (replace_key_set=True), then we assume they will
        # all have been handled by *this* reduce thread, and hence be in the cache already
        # So we only need key select queries if replace_key_set=False
        else:
            key_select_query = select([self.annotation_key_class.id])\
                                .where(self.annotation_key_class.name == bindparam('name'))
            if key_group is not None:
                key_select_query = key_select_query.where(self.annotation_key_class.group == key_group)

        anno_insert_query = self.annotation_class.__table__.insert()

        # Check if the AnnotationKey already exists, and gets its id
        key_id = None
        if key_name in self.key_cache:
            key_id = self.key_cache[key_name]
        else:
            key_args = {'name': key_name, 'group': key_group} if key_group else {'name': key_name}

            # If we are replacing the AnnotationKeys (replace_key_set=True), then we assume they will
            # all have been handled by *this* reduce thread, and hence be in the cache already
            if not replace_key_set:
                key_id = self.session.execute(key_select_query, key_args).first()

            # Key not in cache but exists in DB; add to cache
            if key_id is not None:
                key_id                   = key_id[0]
                self.key_cache[key_name] = key_id

            # Key not in cache or DB; add to both if create_new_keyset = True
            elif replace_key_set:
                key_id   = self.session.execute(key_insert_query, key_args).inserted_primary_key[0]
                self.key_cache[key_name] = key_id

        # If AnnotationKey does not exist and create_new_keyset = False, skip
        if key_id is not None:

            # Updates the Annotation, assuming one might already exist, if try_update = True
            if not clear:
                res = self.session.execute(anno_update_query, {'cid': cid, 'kid': key_id, 'value': value})

            # If Annotation does not exist, insert
            if (clear or res.rowcount == 0) and value != 0:
                self.session.execute(anno_insert_query, {'candidate_id': cid, 'key_id': key_id, 'value': value})


def load_matrix(matrix_class, annotation_key_class, annotation_class, session,
    split=0, cids_query=None, key_group=0, key_names=None, zero_one=False,
    load_as_array=False):
    """
    Returns the annotations corresponding to a split of candidates with N members
    and an AnnotationKey group with M distinct keys as an N x M CSR sparse matrix.
    """
    cid_query = cids_query or session.query(Candidate.id)\
                                     .filter(Candidate.split == split)
    cid_query = cid_query.order_by(Candidate.id)

    keys_query = session.query(annotation_key_class.id)
    keys_query = keys_query.filter(annotation_key_class.group == key_group)
    if key_names is not None:
        keys_query = keys_query.filter(annotation_key_class.name.in_(frozenset(key_names)))
    keys_query = keys_query.order_by(annotation_key_class.id)

    # First, we query to construct the row index map
    cid_to_row = {}
    row_to_cid = {}
    for cid, in cid_query.all():
        if cid not in cid_to_row:
            j = len(cid_to_row)

            # Create both mappings
            cid_to_row[cid] = j
            row_to_cid[j]   = cid

    # Second, we query to construct the column index map
    kid_to_col = {}
    col_to_kid = {}
    for kid, in keys_query.all():
        if kid not in kid_to_col:
            j = len(kid_to_col)

            # Create both mappings
            kid_to_col[kid] = j
            col_to_kid[j]   = kid

    # Create sparse matrix in COO format for incremental construction
    row = []
    columns = []
    data = []

    # Rely on the core for fast iteration
    annot_select_query = annotation_class.__table__.select()

    # Iteratively construct row index and output sparse matrix
    # Cycles through the entire table to load the data.
    # Perfornamce may slow down based on table size; however, negligible since 
    # it takes 8min to go throuh 245M rows (pretty fast).
    for res in session.execute(annot_select_query):
        # NOTE: The order of return seems to be switched in Python 3???
        # Either way, make sure the order is set here explicitly!
        cid, kid, val = res.candidate_id, res.key_id, res.value

        if cid in cid_to_row and kid in kid_to_col:

            # Optionally restricts val range to {0,1}, mapping -1 -> 0
            if zero_one:
                val = 1 if val == 1 else 0
            row.append(cid_to_row[cid])
            columns.append(kid_to_col[kid])
            data.append(int(val))

    X = sparse.coo_matrix((data, (row, columns)), shape=(len(cid_to_row), len(kid_to_col)))

    # Return as an AnnotationMatrix
    Xr = matrix_class(X, candidate_index=cid_to_row, row_index=row_to_cid,
            annotation_key_cls=annotation_key_class, key_index=kid_to_col, 
            col_index=col_to_kid)
    return np.squeeze(Xr.toarray()) if load_as_array else Xr


def load_label_matrix(session, **kwargs):
    return load_matrix(csr_LabelMatrix, LabelKey, Label, session, **kwargs)


def load_feature_matrix(session, **kwargs):
    return load_matrix(csr_AnnotationMatrix, FeatureKey, Feature, session, **kwargs)


def load_gold_labels(session, annotator_name, **kwargs):
    return load_matrix(csr_LabelMatrix, GoldLabelKey, GoldLabel, session, key_names=[annotator_name], **kwargs)


class LabelAnnotator(Annotator):
    """Apply labeling functions to the candidates, generating Label annotations

    :param lfs: A _list_ of labeling functions (LFs)
    """
    def __init__(self, lfs=None, label_generator=None):
        if lfs is not None:
            labels = lambda c : [(lf.__name__, lf(c)) for lf in lfs]
        elif label_generator is not None:
            labels = lambda c : label_generator(c)
        else:
            raise ValueError("Must provide lfs or label_generator kwarg.")

        # Convert lfs to a generator function
        # In particular, catch verbose values and convert to integer ones
        def f_gen(c):
            for lf_key, label in labels(c):
                # Note: We assume if the LF output is an int, it is already
                # mapped correctly
                if isinstance(label, int):
                    yield lf_key, label
                # None is a protected LF output value corresponding to 0,
                # representing LF abstaining
                elif label is None:
                    yield lf_key, 0
                elif label in c.values:
                    if c.cardinality > 2:
                        yield lf_key, c.values.index(label) + 1
                    # Note: Would be nice to not special-case here, but for
                    # consistency we leave binary LF range as {-1,0,1}
                    else:
                        val = 1 if c.values.index(label) == 0 else -1
                        yield lf_key, val
                else:
                    raise ValueError("""
                        Unable to parse label with value %s
                        for candidate with values %s""" % (label, c.values))

        super(LabelAnnotator, self).__init__(Label, LabelKey, f_gen)

    def load_matrix(self, session, **kwargs):
        return load_label_matrix(session, **kwargs)


class FeatureAnnotator(Annotator):
    """Apply feature generators to the candidates, generating Feature annotations"""
    def __init__(self, f):
        super(FeatureAnnotator, self).__init__(Feature, FeatureKey, f)

    def load_matrix(self, session, **kwargs):
        return load_feature_matrix(session, **kwargs)


def save_marginals(session, X, marginals, training=True):
    """Save marginal probabilities for a set of Candidates to db.

    :param X: Either an M x N csr_AnnotationMatrix-class matrix, where M
        is number of candidates, N number of LFs/features; OR a list of
        arbitrary objects with candidate ids accessible via a .id attrib
    :param marginals: A dense M x K matrix of marginal probabilities, where
        K is the cardinality of the candidates, OR a M-dim list/array if K=2.
    :param training: If True, these are training marginals / labels; else they
        are saved as end model predictions.

    Note: The marginals for k=0 are not stored, only for k = 1,...,K
    """
    # Make sure that we are working with a numpy array
    try:
        shape = marginals.shape
    except:
        marginals = np.array(marginals)
        shape = marginals.shape

    # Handle binary input as M x 1-dim array; assume elements represent
    # poksitive (k=1) class values
    if len(shape) == 1:
        marginals = np.vstack([1-marginals, marginals]).T

    # Only add values for classes k=1,...,K
    marginal_tuples = []
    for i in range(shape[0]):
        for k in range(1, shape[1] if len(shape) > 1 else 2):
            if marginals[i, k] > 0:
                marginal_tuples.append((i, k, marginals[i, k]))

    # NOTE: This will delete all existing marginals of type `training`
    session.query(Marginal).filter(Marginal.training == training).\
        delete(synchronize_session='fetch')

    # Prepare bulk INSERT query
    q = Marginal.__table__.insert()

    # Check whether X is an AnnotationMatrix or not
    anno_matrix = isinstance(X, csr_AnnotationMatrix)
    if not anno_matrix:
        X = list(X)

    # Prepare values
    insert_vals = []
    for i, k, p in marginal_tuples:
        cid = X.get_candidate(session, i).id if anno_matrix else X[i].id
        insert_vals.append({
            'candidate_id': cid,
            'training': training,
            'value': k,
            # We cast p in case its a numpy type, which psycopg2 does not handle
            'probability': float(p)
        })

    # Execute update
    session.execute(q, insert_vals)
    session.commit()
    print("Saved %s marginals" % len(marginals))


def load_marginals(session, X=None, split=0, cids_query=None, training=True):
    """Load the marginal probs. for a given split of Candidates"""
    # For candidate ids subquery
    cids_query = cids_query or session.query(Candidate.id) \
        .filter(Candidate.split == split)
    # Ensure ordering by CID
    cids_query = cids_query.order_by(Candidate.id)
    cids_sub_query = cids_query.subquery('cids')

    # Load marginal tuples from db
    marginal_tuples = session.query(Marginal.candidate_id, Marginal.value, 
        Marginal.probability) \
        .filter(Marginal.candidate_id == cids_sub_query.c.id) \
        .filter(Marginal.training == training) \
        .all()

    # If an AnnotationMatrix or list of candidates X is provided, we make sure
    # that the returned marginals are collated with X.
    if X is not None:
        # For now, handle feature matrix vs. list of objects with try / except
        # Handle AnnotationMatrix
        try:
            cardinality = X.get_candidate(session, 0).cardinality
            marginals = np.zeros((X.shape[0], cardinality))
            cid_map = X.candidate_index
        
        # Handle list of Candidates
        except:
            cardinality = X[0].cardinality
            marginals = np.zeros((len(X), cardinality))
            cid_map = dict([(x.id, i) for i, x in enumerate(X)])
    
    # Otherwise if X is not provided, we sort by candidate id, using the
    # cids_query from above
    else:
        cardinality = session.query(Candidate) \
            .get(marginal_tuples[0][0]).cardinality
        marginals = np.zeros((cids_query.count(), cardinality))
        cid_map = dict([(cid, i) for i, (cid,) in enumerate(cids_query.all())])

    # Assemble the marginals matrix according to the candidate index of X
    for cid, k, p in marginal_tuples:
        marginals[cid_map[cid], k] = p

    # Add first column if k > 2, else ravel
    if cardinality > 2:
        row_sums = marginals.sum(axis=1)
        for i in range(marginals.shape[0]):
            marginals[i, 0] = 1 - row_sums[i]
    else:
        marginals = np.ravel(marginals[:, 1])
    return marginals



from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()
from builtins import *
import logging
from multiprocessing import Process, JoinableQueue
from queue import Empty
from .models.meta import new_sessionmaker, snorkel_conn_string
from .utils import ProgressBar


QUEUE_TIMEOUT = 3


class UDFRunner(object):
    """Class to run UDFs in parallel using simple queue-based multiprocessing setup"""
    def __init__(self, udf_class, **udf_init_kwargs):
        self.logger = logging.getLogger(__name__)
        self.udf_class       = udf_class
        self.udf_init_kwargs = udf_init_kwargs
        self.udfs            = []

        if hasattr(self.udf_class, 'reduce'):
            self.reducer = self.udf_class(**self.udf_init_kwargs)
        else:
            self.reducer = None

    def apply(self, xs, clear=True, parallelism=None, progress_bar=True, count=None, **kwargs):
        """
        Apply the given UDF to the set of objects xs, either single or multi-threaded,
        and optionally calling clear() first.
        """
        # Clear everything downstream of this UDF if requested
        if clear:
            self.logger.info("Clearing existing...")
            SnorkelSession = new_sessionmaker()
            session = SnorkelSession()
            self.clear(session, **kwargs)
            session.commit()
            session.close()

        # Execute the UDF
        self.logger.info("Running UDF...")
        if parallelism is None or parallelism < 2:
            self.apply_st(xs, progress_bar, clear=clear, count=count, **kwargs)
        else:
            self.apply_mt(xs, parallelism, clear=clear, **kwargs)

    def clear(self, session, **kwargs):
        raise NotImplementedError()

    def apply_st(self, xs, progress_bar, count, **kwargs):
        """Run the UDF single-threaded, optionally with progress bar"""
        udf = self.udf_class(**self.udf_init_kwargs)

        # Set up ProgressBar if possible
        pb = None
        if progress_bar and hasattr(xs, '__len__') or count is not None:
            n = count if count is not None else len(xs)
            pb = ProgressBar(n)

        # Run single-thread
        for i, x in enumerate(xs):
            if pb:
                pb.bar(i)

            # Apply UDF and add results to the session
            for y in udf.apply(x, **kwargs):

                # Uf UDF has a reduce step, this will take care of the insert; else add to session
                if hasattr(self.udf_class, 'reduce'):
                    udf.reduce(y, **kwargs)
                else:
                    udf.session.add(y)

        # Commit session and close progress bar if applicable
        udf.session.commit()
        if pb:
            pb.close()

    def apply_mt(self, xs, parallelism, **kwargs):
        """Run the UDF multi-threaded using python multiprocessing"""
        if snorkel_conn_string.startswith('sqlite'):
            raise ValueError('Multiprocessing with SQLite is not supported. Please use a different database backend,'
                             ' such as PostgreSQL.')

        # Fill a JoinableQueue with input objects
        in_queue = JoinableQueue()
        for x in xs:
            in_queue.put(x)

        # If the UDF has a reduce step, we collect the output of apply in a Queue
        out_queue = None
        if hasattr(self.udf_class, 'reduce'):
            out_queue = JoinableQueue()

        # Start UDF Processes
        for i in range(parallelism):
            udf              = self.udf_class(in_queue=in_queue, out_queue=out_queue, worker_id = i, **self.udf_init_kwargs)
            udf.apply_kwargs = kwargs
            self.udfs.append(udf)

        # Start the UDF processes, and then join on their completion
        for udf in self.udfs:
            udf.start()

        # If there is a reduce step, do now on this thread
        if hasattr(self.udf_class, 'reduce'):
            while any([udf.is_alive() for udf in self.udfs]):
                while True:
                    try:
                        y = out_queue.get(True, QUEUE_TIMEOUT)
                        self.reducer.reduce(y, **kwargs)
                        out_queue.task_done()
                    except Empty:
                        break
                self.reducer.session.commit()
            self.reducer.session.close()

        # Otherwise just join on the UDF.apply actions
        else:
            for i, udf in enumerate(self.udfs):
                udf.join()

        # Terminate and flush the processes
        for udf in self.udfs:
            udf.terminate()
        self.udfs = []


class UDF(Process):
    def __init__(self, in_queue=None, out_queue=None, worker_id=0):
        """
        in_queue: A Queue of input objects to process; primarily for running in parallel
        """
        Process.__init__(self)
        self.daemon       = True
        self.in_queue     = in_queue
        self.out_queue    = out_queue
        self.worker_id    = worker_id

        # Each UDF starts its own Engine
        # See http://docs.sqlalchemy.org/en/latest/core/pooling.html#using-connection-pools-with-multiprocessing
        SnorkelSession = new_sessionmaker()
        self.session   = SnorkelSession()

        # We use a workaround to pass in the apply kwargs
        self.apply_kwargs = {}

    def run(self):
        """
        This method is called when the UDF is run as a Process in a multiprocess setting
        The basic routine is: get from JoinableQueue, apply, put / add outputs, loop
        """
        while True:
            try:
                x = self.in_queue.get(True, QUEUE_TIMEOUT)
                for y in self.apply(x, **self.apply_kwargs):

                    # If an out_queue is provided, add to that, else add to session
                    if self.out_queue is not None:
                        self.out_queue.put(y, True, QUEUE_TIMEOUT)
                    else:
                        self.session.add(y)
                self.in_queue.task_done()
            except Empty:
                break
        self.session.commit()
        self.session.close()

    def apply(self, x, **kwargs):
        """This function takes in an object, and returns a generator / set / list"""
        raise NotImplementedError()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import *

import numpy as np
from .utils import MentionScorer
from ..annotations import save_marginals


class Classifier(object):
    """Simple abstract base class for a probabilistic classifier."""

    # Set this class variable to True if train, marginals, predict, and score,
    # take a list of @Candidates as the first argument X;
    # otherwise assume X is an AnnotationMatrix
    representation = False

    def __init__(self, cardinality=2, name=None):
        self.name = name or self.__class__.__name__
        self.cardinality = cardinality

    def marginals(self, X, batch_size=None, **kwargs):
        raise NotImplementedError()

    def save_marginals(self, session, X, training=False):
        """Save the predicted marginal probabilities for the Candidates X."""
        save_marginals(session, X, self.marginals(X), training=training)

    def predictions(self, X, b=0.5, batch_size=None):
        """Return numpy array of elements in {-1,0,1}
        based on predicted marginal probabilities.
        """
        if self.cardinality > 2:
            return self.marginals(X, batch_size=batch_size).argmax(axis=1) + 1
        else:
            return np.array([1 if p > b else -1 if p < b else 0 
                for p in self.marginals(X, batch_size=batch_size)])

    def score(self, X_test, Y_test, b=0.5, set_unlabeled_as_neg=True, beta=1,
        batch_size=None):
        """
        Returns the summary scores:
            * For binary: precision, recall, F-beta score
            * For categorical: accuracy
        
        :param X_test: The input test candidates, as a list or annotation matrix
        :param Y_test: The input test labels, as a list or annotation matrix
        :param b: Decision boundary *for binary setting only*
        :param set_unlabeled_as_neg: Whether to map 0 labels -> -1, 
            *binary setting.*
        :param beta: For F-beta score; by default beta = 1 => F-1 score.

        Note: Unlike in self.error_analysis, this method assumes X_test and
        Y_test are properly collated!
        """
        predictions = self.predictions(X_test, b=b, batch_size=batch_size)

        # Convert Y_test to dense numpy array
        try:
            Y_test = np.array(Y_test.todense()).reshape(-1)
        except:
            Y_test = np.array(Y_test)

        # Compute accuracy for categorical, or P/R/F1 for binary settings
        if self.cardinality > 2:
            # Compute and return accuracy
            correct = np.where([predictions == Y_test])[0].shape[0]
            return correct / float(Y_test.shape[0])
        else:
            # Either remap or filter out unlabeled (0-valued) test labels
            if set_unlabeled_as_neg:
                Y_test[Y_test == 0] = -1
            else:
                predictions = predictions[Y_test != 0]
                Y_test = Y_test[Y_test != 0]
            
            # Compute and return precision, recall, and F1 score
            tp = (0.5 * (predictions * Y_test + 1))[predictions == 1].sum()
            pred_pos = predictions[predictions == 1].sum()
            p = tp / float(pred_pos) if pred_pos > 0 else 0.0
            pos = Y_test[Y_test == 1].sum()
            r = tp / float(pos) if pos > 0 else 0.0

            # Compute general F-beta score
            if p + r > 0:
                f_beta = (1 + beta**2) * ((p * r) / (((beta**2) * p) + r))
            else:
                f_beta = 0.0
            return p, r, f_beta

    def error_analysis(self, session, X_test, Y_test, 
        gold_candidate_set=None, b=0.5, set_unlabeled_as_neg=True, display=True,
        scorer=MentionScorer, **kwargs):
        """
        Prints full score analysis using the Scorer class, and then returns the
        a tuple of sets conatining the test candidates bucketed for error 
        analysis, i.e.:
            * For binary: TP, FP, TN, FN
            * For categorical: correct, incorrect
        
        :param X_test: The input test candidates, as a list or annotation matrix
        :param Y_test: The input test labels, as a list or annotation matrix
        :param gold_candidate_set: Full set of TPs in the test set
        :param b: Decision boundary *for binary setting only*
        :param set_unlabeled_as_neg: Whether to map 0 labels -> -1, *binary setting*
        :param display: Print score report
        :param scorer: The Scorer sub-class to use
        """
        # Compute the marginals
        test_marginals = self.marginals(X_test, **kwargs)

        # Get the test candidates
        test_candidates = [
            X_test.get_candidate(session, i) for i in range(X_test.shape[0])
        ] if not self.representation else X_test

        # Initialize and return scorer
        s = scorer(test_candidates, Y_test, gold_candidate_set)          
        return s.score(test_marginals, train_marginals=None, b=b,
            display=display, set_unlabeled_as_neg=set_unlabeled_as_neg)

    def _preprocess_data(self, X):
        """Generic preprocessing subclass; may be called by external methods."""
        return X

    def save(self):
        raise NotImplementedError()

    def load(self):
        raise NotImplementedError()


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import *
from future.utils import iteritems

from collections import defaultdict
from copy import deepcopy
from itertools import product
import re
from sqlalchemy.sql import select

from .models import Candidate, TemporarySpan, Sentence
from .udf import UDF, UDFRunner

QUEUE_COLLECT_TIMEOUT = 5


class CandidateExtractor(UDFRunner):
    """
    An operator to extract Candidate objects from a Context.

    :param candidate_class: The type of relation to extract, defined using
                            :func:`snorkel.models.candidate_subclass <snorkel.models.candidate.candidate_subclass>`
    :param cspaces: one or list of :class:`CandidateSpace` objects, one for each relation argument. Defines space of
                    Contexts to consider
    :param matchers: one or list of :class:`snorkel.matchers.Matcher` objects, one for each relation argument. Only tuples of
                     Contexts for which each element is accepted by the corresponding Matcher will be returned as Candidates
    :param candidate_filter: an optional function for filtering out candidates which returns a Boolean expressing whether or not
                      the candidate should be instantiated.
    :param self_relations: Boolean indicating whether to extract Candidates that relate the same context.
                           Only applies to binary relations. Default is False.
    :param nested_relations: Boolean indicating whether to extract Candidates that relate one Context with another
                             that contains it. Only applies to binary relations. Default is False.
    :param symmetric_relations: Boolean indicating whether to extract symmetric Candidates, i.e., rel(A,B) and rel(B,A),
                                where A and B are Contexts. Only applies to binary relations. Default is False.
    """
    def __init__(self, candidate_class, cspaces, matchers, candidate_filter=None, self_relations=False, nested_relations=False, symmetric_relations=False):
        super(CandidateExtractor, self).__init__(CandidateExtractorUDF,
                                                 candidate_class=candidate_class,
                                                 cspaces=cspaces,
                                                 matchers=matchers,
                                                 candidate_filter=candidate_filter,
                                                 self_relations=self_relations,
                                                 nested_relations=nested_relations,
                                                 symmetric_relations=symmetric_relations)

    def apply(self, xs, split=0, **kwargs):
        super(CandidateExtractor, self).apply(xs, split=split, **kwargs)

    def clear(self, session, split, **kwargs):
        session.query(Candidate).filter(Candidate.split == split).delete()


class CandidateExtractorUDF(UDF):
    def __init__(self, candidate_class, cspaces, matchers, candidate_filter, self_relations, nested_relations, symmetric_relations, **kwargs):
        self.candidate_class     = candidate_class
        # Note: isinstance is the way to check types -- not type(x) in [...]!
        self.candidate_spaces    = cspaces if isinstance(cspaces, (list, tuple)) else [cspaces]
        self.matchers            = matchers if isinstance(matchers, (list, tuple)) else [matchers]
        self.candidate_filter    = candidate_filter
        self.nested_relations    = nested_relations
        self.self_relations      = self_relations
        self.symmetric_relations = symmetric_relations

        # Check that arity is same
        if len(self.candidate_spaces) != len(self.matchers):
            raise ValueError("Mismatched arity of candidate space and matcher.")
        else:
            self.arity = len(self.candidate_spaces)

        # Make sure the candidate spaces are different so generators aren't expended!
        self.candidate_spaces = list(map(deepcopy, self.candidate_spaces))

        # Preallocates internal data structures
        self.child_context_sets = [None] * self.arity
        for i in range(self.arity):
            self.child_context_sets[i] = set()

        super(CandidateExtractorUDF, self).__init__(**kwargs)

    def apply(self, context, clear, split, **kwargs):
        # Generate TemporaryContexts that are children of the context using the candidate_space and filtered
        # by the Matcher
        for i in range(self.arity):
            self.child_context_sets[i].clear()
            for tc in self.matchers[i].apply(self.candidate_spaces[i].apply(context)):
                tc.load_id_or_insert(self.session)
                self.child_context_sets[i].add(tc)

        # Generates and persists candidates
        extracted = set()
        candidate_args = {'split': split}
        for args in product(*[enumerate(child_contexts) for child_contexts in self.child_context_sets]):

            # Apply candidate_filter if one was given
            # Accepts a tuple of Context objects (e.g., (Span, Span))
            # (candidate_filter returns whether or not proposed candidate passes throttling condition)
            if self.candidate_filter:
                if not self.candidate_filter(tuple(args[i][1] for i in range(self.arity))):
                    continue

            # TODO: Make this work for higher-order relations
            if self.arity == 2:
                ai, a = args[0]
                bi, b = args[1]

                # Check for self-joins, "nested" joins (joins from span to its subspan), and flipped duplicate
                # "symmetric" relations. For symmetric relations, if mentions are of the same type, maintain
                # their order in the sentence.
                if not self.self_relations and a == b:
                    continue
                elif not self.nested_relations and (a in b or b in a):
                    continue
                elif not self.symmetric_relations and ((b, a) in extracted or
                    (self.matchers[0] == self.matchers[1] and a.char_start > b.char_start)):
                    continue

                # Keep track of extracted
                extracted.add((a,b))

            # Assemble candidate arguments
            for i, arg_name in enumerate(self.candidate_class.__argnames__):
                candidate_args[arg_name + '_id'] = args[i][1].id

            # Checking for existence
            if not clear:
                q = select([self.candidate_class.id])
                for key, value in iteritems(candidate_args):
                    q = q.where(getattr(self.candidate_class, key) == value)
                candidate_id = self.session.execute(q).first()
                if candidate_id is not None:
                    continue

            # Add Candidate to session
            yield self.candidate_class(**candidate_args)


class CandidateSpace(object):
    """
    Defines the **space** of candidate objects
    Calling _apply(x)_ given an object _x_ returns a generator over candidates in _x_.
    """
    def __init__(self):
        pass

    def apply(self, x):
        raise NotImplementedError()


class Ngrams(CandidateSpace):
    """
    Defines the space of candidates as all n-grams (n <= n_max) in a Sentence _x_,
    indexing by **character offset**.
    """
    def __init__(self, n_max=5, split_tokens=('-', '/')):
        CandidateSpace.__init__(self)
        self.n_max     = n_max
        self.split_rgx = r'('+r'|'.join(split_tokens)+r')' if split_tokens and len(split_tokens) > 0 else None

    def apply(self, context):

        # These are the character offset--**relative to the sentence start**--for each _token_
        offsets = context.char_offsets

        # Loop over all n-grams in **reverse** order (to facilitate longest-match semantics)
        L    = len(offsets)
        seen = set()
        for l in range(1, self.n_max+1)[::-1]:
            for i in range(L-l+1):
                w     = context.words[i+l-1]
                start = offsets[i]
                end   = offsets[i+l-1] + len(w) - 1
                ts    = TemporarySpan(char_start=start, char_end=end, sentence=context)
                if ts not in seen:
                    seen.add(ts)
                    yield ts

                # Check for split
                # NOTE: For simplicity, we only split single tokens right now!
                if l == 1 and self.split_rgx is not None and end - start > 0:
                    m = re.search(self.split_rgx, context.text[start-offsets[0]:end-offsets[0]+1])
                    if m is not None and l < self.n_max + 1:
                        ts1 = TemporarySpan(char_start=start, char_end=start + m.start(1) - 1, sentence=context)
                        if ts1 not in seen:
                            seen.add(ts1)
                            yield ts
                        ts2 = TemporarySpan(char_start=start + m.end(1), char_end=end, sentence=context)
                        if ts2 not in seen:
                            seen.add(ts2)
                            yield ts2


class PretaggedCandidateExtractor(UDFRunner):
    """UDFRunner for PretaggedCandidateExtractorUDF"""
    def __init__(self, candidate_class, entity_types, self_relations=False,
     nested_relations=False, symmetric_relations=True, entity_sep='~@~'):
        super(PretaggedCandidateExtractor, self).__init__(
            PretaggedCandidateExtractorUDF, candidate_class=candidate_class,
            entity_types=entity_types, self_relations=self_relations,
            nested_relations=nested_relations, entity_sep=entity_sep,
            symmetric_relations=symmetric_relations,
        )

    def apply(self, xs, split=0, **kwargs):
        super(PretaggedCandidateExtractor, self).apply(xs, split=split, **kwargs)

    def clear(self, session, split, **kwargs):
        session.query(Candidate).filter(Candidate.split == split).delete()


class PretaggedCandidateExtractorUDF(UDF):
    """
    An extractor for Sentences with entities pre-tagged, and stored in the entity_types and entity_cids
    fields.
    """
    def __init__(self, candidate_class, entity_types, self_relations=False, nested_relations=False, symmetric_relations=False, entity_sep='~@~', **kwargs):
        self.candidate_class     = candidate_class
        self.entity_types        = entity_types
        self.arity               = len(entity_types)
        self.self_relations      = self_relations
        self.nested_relations    = nested_relations
        self.symmetric_relations = symmetric_relations
        self.entity_sep          = entity_sep

        super(PretaggedCandidateExtractorUDF, self).__init__(**kwargs)

    def apply(self, context, clear, split, check_for_existing=True, **kwargs):
        """Extract Candidates from a Context"""
        # For now, just handle Sentences
        if not isinstance(context, Sentence):
            raise NotImplementedError("%s is currently only implemented for Sentence contexts." % self.__name__)

        # Do a first pass to collect all mentions by entity type / cid
        entity_idxs = dict((et, defaultdict(list)) for et in set(self.entity_types))
        L = len(context.words)
        for i in range(L):
            if context.entity_types[i] is not None:
                ets  = context.entity_types[i].split(self.entity_sep)
                cids = context.entity_cids[i].split(self.entity_sep)
                for et, cid in zip(ets, cids):
                    if et in entity_idxs:
                        entity_idxs[et][cid].append(i)

        # Form entity Spans
        entity_spans = defaultdict(list)
        entity_cids  = {}
        for et, cid_idxs in iteritems(entity_idxs):
            for cid, idxs in iteritems(entity_idxs[et]):
                while len(idxs) > 0:
                    i          = idxs.pop(0)
                    char_start = context.char_offsets[i]
                    char_end   = char_start + len(context.words[i]) - 1
                    while len(idxs) > 0 and idxs[0] == i + 1:
                        i        = idxs.pop(0)
                        char_end = context.char_offsets[i] + len(context.words[i]) - 1

                    # Insert / load temporary span, also store map to entity CID
                    tc = TemporarySpan(char_start=char_start, char_end=char_end, sentence=context)
                    tc.load_id_or_insert(self.session)
                    entity_cids[tc.id] = cid
                    entity_spans[et].append(tc)

        # Generates and persists candidates
        candidate_args = {'split' : split}
        for args in product(*[enumerate(entity_spans[et]) for et in self.entity_types]):

            # TODO: Make this work for higher-order relations
            if self.arity == 2:
                ai, a = args[0]
                bi, b = args[1]

                # Check for self-joins, "nested" joins (joins from span to its subspan), and flipped duplicate
                # "symmetric" relations
                if not self.self_relations and a == b:
                    continue
                elif not self.nested_relations and (a in b or b in a):
                    continue
                elif not self.symmetric_relations and ai > bi:
                    continue

            # Assemble candidate arguments
            for i, arg_name in enumerate(self.candidate_class.__argnames__):
                candidate_args[arg_name + '_id'] = args[i][1].id
                candidate_args[arg_name + '_cid'] = entity_cids[args[i][1].id]

            # Checking for existence
            if check_for_existing:
                q = select([self.candidate_class.id])
                for key, value in iteritems(candidate_args):
                    q = q.where(getattr(self.candidate_class, key) == value)
                candidate_id = self.session.execute(q).first()
                if candidate_id is not None:
                    continue

            # Add Candidate to session
            yield self.candidate_class(**candidate_args)



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from builtins import str
from builtins import range
from collections import defaultdict
from itertools import chain
from lxml import etree
from lxml.html import fromstring
import numpy as np

from fonduer.models import Phrase
from fonduer.snorkel.candidates import Ngrams
from fonduer.snorkel.models.context import TemporarySpan
from fonduer.snorkel.utils import tokens_to_ngrams
from fonduer.utils_table import (min_row_diff, min_col_diff, is_row_aligned,
                                 is_col_aligned, is_axis_aligned)
from fonduer.utils_visual import (
    bbox_from_span, bbox_from_phrase, bbox_horz_aligned, bbox_vert_aligned,
    bbox_vert_aligned_left, bbox_vert_aligned_right, bbox_vert_aligned_center)


def get_between_ngrams(c, attrib='words', n_min=1, n_max=1, lower=True):
    """Return the ngrams _between_ two unary Spans of a binary-Span Candidate.

    Get the ngrams _between_ two unary Spans of a binary-Span Candidate, where
    both share the same sentence Context.

    :param c: The binary-Span Candidate to evaluate.
    :param attrib: The token attribute type (e.g. words, lemmas, poses)
    :param n_min: The minimum n of the ngrams that should be returned
    :param n_max: The maximum n of the ngrams that should be returned
    :param lower: If 'True', all ngrams will be returned in lower case
    :rtype: a _generator_ of ngrams
    """
    if len(c) != 2:
        raise ValueError("Only applicable to binary Candidates")
    span0 = c[0]
    span1 = c[1]
    if span0.sentence != span1.sentence:
        raise ValueError("Only applicable to Candidates where both spans are \
                          from the same immediate Context.")
    distance = abs(span0.get_word_start() - span1.get_word_start())
    if span0.get_word_start() < span1.get_word_start():
        for ngram in get_right_ngrams(
                span0,
                window=distance - 1,
                attrib=attrib,
                n_min=n_min,
                n_max=n_max,
                lower=lower):
            yield ngram
    else:  # span0.get_word_start() > span1.get_word_start()
        for ngram in get_left_ngrams(
                span1,
                window=distance - 1,
                attrib=attrib,
                n_min=n_min,
                n_max=n_max,
                lower=lower):
            yield ngram


def get_left_ngrams(span,
                    window=3,
                    attrib='words',
                    n_min=1,
                    n_max=1,
                    lower=True):
    """Get the ngrams within a window to the _left_ of the Candidate from its sentence Context.

    For higher-arity Candidates, defaults to the _first_ argument.

    :param span: The Span to evaluate. If a candidate is given, default to its first Span.
    :param window: The number of tokens to the left of the first argument to return
    :param attrib: The token attribute type (e.g. words, lemmas, poses)
    :param n_min: The minimum n of the ngrams that should be returned
    :param n_max: The maximum n of the ngrams that should be returned
    :param lower: If True, all ngrams will be returned in lower case
    :rtype: a _generator_ of ngrams
    """
    span = span if isinstance(span,
                              TemporarySpan) else span[0]  # get first Span
    i = span.get_word_start()
    for ngram in tokens_to_ngrams(
            getattr(span.sentence, attrib)[max(0, i - window):i],
            n_min=n_min,
            n_max=n_max,
            lower=lower):
        yield ngram


def get_right_ngrams(span,
                     window=3,
                     attrib='words',
                     n_min=1,
                     n_max=1,
                     lower=True):
    """Get the ngrams within a window to the _right_ of the Candidate from its sentence Context.

    For higher-arity Candidates, defaults to the _last_ argument.

    :param span: The Span to evaluate. If a candidate is given, default to its last Span.
    :param window: The number of tokens to the left of the first argument to return
    :param attrib: The token attribute type (e.g. words, lemmas, poses)
    :param n_min: The minimum n of the ngrams that should be returned
    :param n_max: The maximum n of the ngrams that should be returned
    :param lower: If True, all ngrams will be returned in lower case
    :rtype: a _generator_ of ngrams
    """
    span = span if isinstance(span,
                              TemporarySpan) else span[-1]  # get last Span
    i = span.get_word_end()
    for ngram in tokens_to_ngrams(
            getattr(span.sentence, attrib)[i + 1:i + 1 + window],
            n_min=n_min,
            n_max=n_max,
            lower=lower):
        yield ngram


def get_matches(lf, candidate_set, match_values=[1, -1]):
    """Return a list of candidates that are matched by a particular LF.

    A simple helper function to see how many matches (non-zero by default) an LF gets.
    Returns the matched candidates, which can then be directly put into the Viewer.

    :param lf: The labeling function to apply to the candidate_set
    :param candidate_set: The set of candidates to evaluate
    :param match_values: An option list of the values to consider as matched. [1, -1] by default.
    :rtype: a list of candidates
    """
    matches = []
    for c in candidate_set:
        label = lf(c)
        if label in match_values:
            matches.append(c)
    print(("%s matches") % len(matches))
    return matches


# TABLE LF HELPERS ##########################################################
def same_document(c):
    """Return True if all Spans in the given candidate are from the same Document.

    :param c: The candidate whose Spans are being compared
    :rtype: boolean
    """
    return (all(c[i].sentence.document is not None
                and c[i].sentence.document == c[0].sentence.document
                for i in range(len(c))))


def same_table(c):
    """Return True if all Spans in the given candidate are from the same Table.

    :param c: The candidate whose Spans are being compared
    :rtype: boolean
    """
    return (all(c[i].sentence.is_tabular()
                and c[i].sentence.table == c[0].sentence.table
                for i in range(len(c))))


def same_row(c):
    """Return True if all Spans in the given candidate are from the same Row.

    :param c: The candidate whose Spans are being compared
    :rtype: boolean
    """
    return (same_table(c) and all(
        is_row_aligned(c[i].sentence, c[0].sentence) for i in range(len(c))))


def same_col(c):
    """Return True if all Spans in the given candidate are from the same Col.

    :param c: The candidate whose Spans are being compared
    :rtype: boolean
    """
    return (same_table(c) and all(
        is_col_aligned(c[i].sentence, c[0].sentence) for i in range(len(c))))


def is_tabular_aligned(c):
    """Return True if all Spans in the given candidate are from the same Row or Col.

    :param c: The candidate whose Spans are being compared
    :rtype: boolean
    """
    return (same_table(c) and (is_col_aligned(c[i].sentence, c[0].sentence)
                               or is_row_aligned(c[i].sentence, c[0].sentence)
                               for i in range(len(c))))


def same_cell(c):
    """Return True if all Spans in the given candidate are from the same Cell.

    :param c: The candidate whose Spans are being compared
    :rtype: boolean
    """
    return (all(c[i].sentence.cell is not None
                and c[i].sentence.cell == c[0].sentence.cell
                for i in range(len(c))))


def same_phrase(c):
    """Return True if all Spans in the given candidate are from the same Phrase.

    :param c: The candidate whose Spans are being compared
    :rtype: boolean
    """
    return (all(c[i].sentence is not None and c[i].sentence == c[0].sentence
                for i in range(len(c))))


def get_max_col_num(span):
    """Return the largest column number that a Span occupies.

    :param span: The Span to evaluate. If a candidate is given, default to its last Span.
    :rtype: integer or None
    """
    span = span if isinstance(span, TemporarySpan) else span[-1]
    if span.sentence.is_tabular():
        return span.sentence.cell.col_end
    else:
        return None


def get_min_col_num(span):
    """Return the lowest column number that a Span occupies.

    :param span: The Span to evaluate. If a candidate is given, default to its first Span.
    :rtype: integer or None
    """
    span = span if isinstance(span, TemporarySpan) else span[0]
    if span.sentence.is_tabular():
        return span.sentence.cell.col_start
    else:
        return None


def get_phrase_ngrams(span, attrib='words', n_min=1, n_max=1, lower=True):
    """Get the ngrams that are in the Phrase of the given span, not including itself.

    Note that if a candidate is passed in, all of its Spans will be searched.

    :param span: The Span whose Phrase is being searched
    :param attrib: The token attribute type (e.g. words, lemmas, poses)
    :param n_min: The minimum n of the ngrams that should be returned
    :param n_max: The maximum n of the ngrams that should be returned
    :param lower: If True, all ngrams will be returned in lower case
    :rtype: a _generator_ of ngrams
    """
    spans = [span] if isinstance(span, TemporarySpan) else span.get_contexts()
    for span in spans:
        for ngram in get_left_ngrams(
                span,
                window=100,
                attrib=attrib,
                n_min=n_min,
                n_max=n_max,
                lower=lower):
            yield ngram
        for ngram in get_right_ngrams(
                span,
                window=100,
                attrib=attrib,
                n_min=n_min,
                n_max=n_max,
                lower=lower):
            yield ngram


def get_neighbor_phrase_ngrams(span,
                               d=1,
                               attrib='words',
                               n_min=1,
                               n_max=1,
                               lower=True):
    """Get the ngrams that are in the neighoring Phrases of the given Span.

    Note that if a candidate is passed in, all of its Spans will be searched.

    :param span: The span whose neighbor Phrases are being searched
    :param attrib: The token attribute type (e.g. words, lemmas, poses)
    :param n_min: The minimum n of the ngrams that should be returned
    :param n_max: The maximum n of the ngrams that should be returned
    :param lower: If True, all ngrams will be returned in lower case
    :rtype: a _generator_ of ngrams
    """
    spans = [span] if isinstance(span, TemporarySpan) else span.get_contexts()
    for span in spans:
        for ngram in chain.from_iterable([
                tokens_to_ngrams(
                    getattr(phrase, attrib),
                    n_min=n_min,
                    n_max=n_max,
                    lower=lower) for phrase in span.sentence.document.phrases
                if abs(phrase.phrase_num - span.sentence.phrase_num) <= d
                and phrase != span.sentence
        ]):
            yield ngram


def get_cell_ngrams(span, attrib='words', n_min=1, n_max=1, lower=True):
    """Get the ngrams that are in the Cell of the given span, not including itself.

    Note that if a candidate is passed in, all of its Spans will be searched.

    :param span: The span whose Cell is being searched
    :param attrib: The token attribute type (e.g. words, lemmas, poses)
    :param n_min: The minimum n of the ngrams that should be returned
    :param n_max: The maximum n of the ngrams that should be returned
    :param lower: If True, all ngrams will be returned in lower case
    :rtype: a _generator_ of ngrams
    """
    spans = [span] if isinstance(span, TemporarySpan) else span.get_contexts()
    for span in spans:
        for ngram in get_phrase_ngrams(
                span, attrib=attrib, n_min=n_min, n_max=n_max, lower=lower):
            yield ngram
        if isinstance(span.sentence,
                      Phrase) and span.sentence.cell is not None:
            for ngram in chain.from_iterable([
                    tokens_to_ngrams(
                        getattr(phrase, attrib),
                        n_min=n_min,
                        n_max=n_max,
                        lower=lower) for phrase in span.sentence.cell.phrases
                    if phrase != span.sentence
            ]):
                yield ngram


def get_neighbor_cell_ngrams(span,
                             dist=1,
                             directions=False,
                             attrib='words',
                             n_min=1,
                             n_max=1,
                             lower=True):
    """Get the ngrams from all Cells that are within a given Cell distance in one direction from the given Span.

    Note that if a candidate is passed in, all of its Spans will be searched.
    If `directions=True``, each ngram will be returned with a direction in {'UP', 'DOWN', 'LEFT', 'RIGHT'}.

    :param span: The span whose neighbor Cells are being searched
    :param dist: The Cell distance within which a neighbor Cell must be to be considered
    :param directions: A Boolean expressing whether or not to return the direction of each ngram
    :param attrib: The token attribute type (e.g. words, lemmas, poses)
    :param n_min: The minimum n of the ngrams that should be returned
    :param n_max: The maximum n of the ngrams that should be returned
    :param lower: If True, all ngrams will be returned in lower case
    :rtype: a _generator_ of ngrams (or (ngram, direction) tuples if directions=True)
    """
    # TODO: Fix this to be more efficient (optimize with SQL query)
    spans = [span] if isinstance(span, TemporarySpan) else span.get_contexts()
    for span in spans:
        for ngram in get_phrase_ngrams(
                span, attrib=attrib, n_min=n_min, n_max=n_max, lower=lower):
            yield ngram
        if isinstance(span.sentence,
                      Phrase) and span.sentence.cell is not None:
            root_cell = span.sentence.cell
            for phrase in chain.from_iterable([
                    _get_aligned_phrases(root_cell, 'row'),
                    _get_aligned_phrases(root_cell, 'col')
            ]):
                row_diff = min_row_diff(phrase, root_cell, absolute=False)
                col_diff = min_col_diff(phrase, root_cell, absolute=False)
                if (row_diff or col_diff) and not (
                        row_diff and
                        col_diff) and abs(row_diff) + abs(col_diff) <= dist:
                    if directions:
                        direction = ''
                        if col_diff == 0:
                            if 0 < row_diff and row_diff <= dist:
                                direction = "UP"
                            elif 0 > row_diff and row_diff >= -dist:
                                direction = "DOWN"
                        elif row_diff == 0:
                            if 0 < col_diff and col_diff <= dist:
                                direction = "RIGHT"
                            elif 0 > col_diff and col_diff >= -dist:
                                direction = "LEFT"
                        for ngram in tokens_to_ngrams(
                                getattr(phrase, attrib),
                                n_min=n_min,
                                n_max=n_max,
                                lower=lower):
                            yield (ngram, direction)
                    else:
                        for ngram in tokens_to_ngrams(
                                getattr(phrase, attrib),
                                n_min=n_min,
                                n_max=n_max,
                                lower=lower):
                            yield ngram


def get_row_ngrams(span,
                   attrib='words',
                   n_min=1,
                   n_max=1,
                   spread=[0, 0],
                   lower=True):
    """Get the ngrams from all Cells that are in the same row as the given Span.

    Note that if a candidate is passed in, all of its Spans will be searched.

    :param span: The span whose row Cells are being searched
    :param attrib: The token attribute type (e.g. words, lemmas, poses)
    :param n_min: The minimum n of the ngrams that should be returned
    :param n_max: The maximum n of the ngrams that should be returned
    :param lower: If True, all ngrams will be returned in lower case
    :rtype: a _generator_ of ngrams
    """
    spans = [span] if isinstance(span, TemporarySpan) else span.get_contexts()
    for span in spans:
        for ngram in _get_axis_ngrams(
                span,
                axis='row',
                attrib=attrib,
                n_min=n_min,
                n_max=n_max,
                spread=spread,
                lower=lower):
            yield ngram


def get_col_ngrams(span,
                   attrib='words',
                   n_min=1,
                   n_max=1,
                   spread=[0, 0],
                   lower=True):
    """Get the ngrams from all Cells that are in the same column as the given Span.

    Note that if a candidate is passed in, all of its Spans will be searched.

    :param span: The span whose column Cells are being searched
    :param attrib: The token attribute type (e.g. words, lemmas, poses)
    :param n_min: The minimum n of the ngrams that should be returned
    :param n_max: The maximum n of the ngrams that should be returned
    :param lower: If True, all ngrams will be returned in lower case
    :rtype: a _generator_ of ngrams
    """
    spans = [span] if isinstance(span, TemporarySpan) else span.get_contexts()
    for span in spans:
        for ngram in _get_axis_ngrams(
                span,
                axis='col',
                attrib=attrib,
                n_min=n_min,
                n_max=n_max,
                spread=spread,
                lower=lower):
            yield ngram


def get_aligned_ngrams(span,
                       attrib='words',
                       n_min=1,
                       n_max=1,
                       spread=[0, 0],
                       lower=True):
    """Get the ngrams from all Cells that are in the same row or column as the given Span.

    Note that if a candidate is passed in, all of its Spans will be searched.

    :param span: The span whose row and column Cells are being searched
    :param attrib: The token attribute type (e.g. words, lemmas, poses)
    :param n_min: The minimum n of the ngrams that should be returned
    :param n_max: The maximum n of the ngrams that should be returned
    :param lower: If True, all ngrams will be returned in lower case
    :rtype: a _generator_ of ngrams
    """
    spans = [span] if isinstance(span, TemporarySpan) else span.get_contexts()
    for span in spans:
        for ngram in get_row_ngrams(
                span,
                attrib=attrib,
                n_min=n_min,
                n_max=n_max,
                spread=spread,
                lower=lower):
            yield ngram
        for ngram in get_col_ngrams(
                span,
                attrib=attrib,
                n_min=n_min,
                n_max=n_max,
                spread=spread,
                lower=lower):
            yield ngram


def get_head_ngrams(span,
                    axis=None,
                    attrib='words',
                    n_min=1,
                    n_max=1,
                    lower=True):
    """Get the ngrams from the cell in the head of the row or column.

    More specifically, this returns the ngrams in the leftmost cell in a row and/or the
    ngrams in the topmost cell in the column, depending on the axis parameter.

    Note that if a candidate is passed in, all of its Spans will be searched.

    :param span: The span whose head Cells are being returned
    :param axis: Which axis {'row', 'col'} to search. If None, then both row and col are searched.
    :param attrib: The token attribute type (e.g. words, lemmas, poses)
    :param n_min: The minimum n of the ngrams that should be returned
    :param n_max: The maximum n of the ngrams that should be returned
    :param lower: If True, all ngrams will be returned in lower case
    :rtype: a _generator_ of ngrams
    """
    spans = [span] if isinstance(span, TemporarySpan) else span.get_contexts()
    axes = [axis] if axis else ['row', 'col']
    for span in spans:
        if not span.sentence.cell:
            return
        else:
            for axis in axes:
                if getattr(span.sentence, _other_axis(axis) + '_start') == 0:
                    return
                for phrase in getattr(
                        _get_head_cell(span.sentence.cell, axis), 'phrases',
                    []):
                    for ngram in tokens_to_ngrams(
                            getattr(phrase, attrib),
                            n_min=n_min,
                            n_max=n_max,
                            lower=lower):
                        yield ngram


def _get_head_cell(root_cell, axis):
    other_axis = 'row' if axis == 'col' else 'col'
    aligned_cells = _get_aligned_cells(root_cell, axis)
    return sorted(
        aligned_cells, key=lambda x: getattr(x, other_axis + '_start'))[
            0] if aligned_cells else []


def _get_axis_ngrams(span,
                     axis,
                     attrib='words',
                     n_min=1,
                     n_max=1,
                     spread=[0, 0],
                     lower=True):
    for ngram in get_phrase_ngrams(
            span, attrib=attrib, n_min=n_min, n_max=n_max, lower=lower):
        yield ngram
    if (span.sentence.cell is not None):
        for phrase in _get_aligned_phrases(span.sentence, axis, spread=spread):
            for ngram in tokens_to_ngrams(
                    getattr(phrase, attrib), n_min=n_min, n_max=n_max,
                    lower=lower):
                yield ngram


def _get_aligned_cells(root_cell, axis):
    aligned_cells = [
        cell for cell in root_cell.table.cells
        if is_axis_aligned(root_cell, cell, axis=axis) and cell != root_cell
    ]
    return aligned_cells


def _get_aligned_phrases(root_phrase, axis, spread=[0, 0]):
    return [
        phrase for cell in root_phrase.table.cells
        if is_axis_aligned(root_phrase, cell, axis=axis, spread=spread)
        for phrase in cell.phrases if phrase != root_phrase
    ]


def _other_axis(axis):
    return 'row' if axis == 'col' else 'col'


def is_superset(a, b):
    """Check if a is a superset of b.

    This is typically used to check if ALL of a list of phrases is in the ngrams returned by an lf_helper.

    :param a: A collection of items
    :param b: A collection of items
    :rtype: boolean
    """
    return set(a).issuperset(b)


def overlap(a, b):
    """Check if a overlaps b.

    This is typically used to check if ANY of a list of phrases is in the ngrams returned by an lf_helper.

    :param a: A collection of items
    :param b: A collection of items
    :rtype: boolean
    """
    return not set(a).isdisjoint(b)


############################
# Visual feature helpers
############################
def get_page(span):
    """Return the page number of the given span.

    If a candidate is passed in, this returns the page of its first Span.

    :param span: The Span to get the page number of.
    :rtype: integer
    """
    span = span if isinstance(span, TemporarySpan) else span[0]
    return span.get_attrib_tokens('page')[0]


def is_horz_aligned(c):
    """Return True if all the components of c are horizontally aligned.

    Horizontal alignment means that the bounding boxes of each Span of c shares
    a similar y-axis value in the visual rendering of the document.

    :param c: The candidate to evaluate
    :rtype: boolean
    """
    return (all([
        c[i].sentence.is_visual()
        and bbox_horz_aligned(bbox_from_span(c[i]), bbox_from_span(c[0]))
        for i in range(len(c))
    ]))


def is_vert_aligned(c):
    """Return true if all the components of c are vertically aligned.

    Vertical alignment means that the bounding boxes of each Span of c shares
    a similar x-axis value in the visual rendering of the document.

    :param c: The candidate to evaluate
    :rtype: boolean
    """
    return (all([
        c[i].sentence.is_visual()
        and bbox_vert_aligned(bbox_from_span(c[i]), bbox_from_span(c[0]))
        for i in range(len(c))
    ]))


def is_vert_aligned_left(c):
    """Return true if all the components of c are vertically aligned based on their left border.

    Vertical alignment means that the bounding boxes of each Span of c shares
    a similar x-axis value in the visual rendering of the document. In this function
    the similarity of the x-axis value is based on the left border of their bounding boxes.

    :param c: The candidate to evaluate
    :rtype: boolean
    """
    return (all([
        c[i].sentence.is_visual()
        and bbox_vert_aligned_left(bbox_from_span(c[i]), bbox_from_span(c[0]))
        for i in range(len(c))
    ]))


def is_vert_aligned_right(c):
    """Return true if all the components of c are vertically aligned based on their right border.

    Vertical alignment means that the bounding boxes of each Span of c shares
    a similar x-axis value in the visual rendering of the document. In this function
    the similarity of the x-axis value is based on the right border of their bounding boxes.

    :param c: The candidate to evaluate
    :rtype: boolean
    """
    return (all([
        c[i].sentence.is_visual() and bbox_vert_aligned_right(
            bbox_from_span(c[i]), bbox_from_span(c[0])) for i in range(len(c))
    ]))


def is_vert_aligned_center(c):
    """Return true if all the components of c are vertically aligned based on their left border.

    Vertical alignment means that the bounding boxes of each Span of c shares
    a similar x-axis value in the visual rendering of the document. In this function
    the similarity of the x-axis value is based on the center of their bounding boxes.

    :param c: The candidate to evaluate
    :rtype: boolean
    """
    return (all([
        c[i].sentence.is_visual() and bbox_vert_aligned_center(
            bbox_from_span(c[i]), bbox_from_span(c[0])) for i in range(len(c))
    ]))


def same_page(c):
    """Return true if all the components of c are on the same page of the document.

    Page numbers are based on the PDF rendering of the document. If a PDF file is
    provided, it is used. Otherwise, if only a HTML/XML document is provided, a
    PDF is created and then used to determine the page number of a Span.

    :param c: The candidate to evaluate
    :rtype: boolean
    """
    return (all([
        c[i].sentence.is_visual()
        and bbox_from_span(c[i]).page == bbox_from_span(c[0]).page
        for i in range(len(c))
    ]))


def get_horz_ngrams(span,
                    attrib='words',
                    n_min=1,
                    n_max=1,
                    lower=True,
                    from_phrase=True):
    """Return all ngrams which are visually horizontally aligned with the Span.

    Note that if a candidate is passed in, all of its Spans will be searched.

    :param span: The Span to evaluate
    :param attrib: The token attribute type (e.g. words, lemmas, poses)
    :param n_min: The minimum n of the ngrams that should be returned
    :param n_max: The maximum n of the ngrams that should be returned
    :param lower: If True, all ngrams will be returned in lower case
    :param from_phrase: If True, returns ngrams from any horizontally aligned Phrases,
                        rather than just horizontally aligned ngrams themselves.
    :rtype: a _generator_ of ngrams
    """
    spans = [span] if isinstance(span, TemporarySpan) else span.get_contexts()
    for span in spans:
        for ngram in _get_direction_ngrams('horz', span, attrib, n_min, n_max,
                                           lower, from_phrase):
            yield ngram


def get_vert_ngrams(span,
                    attrib='words',
                    n_min=1,
                    n_max=1,
                    lower=True,
                    from_phrase=True):
    """Return all ngrams which are visually vertivally aligned with the Span.

    Note that if a candidate is passed in, all of its Spans will be searched.

    :param span: The Span to evaluate
    :param attrib: The token attribute type (e.g. words, lemmas, poses)
    :param n_min: The minimum n of the ngrams that should be returned
    :param n_max: The maximum n of the ngrams that should be returned
    :param lower: If True, all ngrams will be returned in lower case
    :param from_phrase: If True, returns ngrams from any horizontally aligned Phrases,
                        rather than just horizontally aligned ngrams themselves.
    :rtype: a _generator_ of ngrams
    """
    spans = [span] if isinstance(span, TemporarySpan) else span.get_contexts()
    for span in spans:
        for ngram in _get_direction_ngrams('vert', span, attrib, n_min, n_max,
                                           lower, from_phrase):
            yield ngram


def _get_direction_ngrams(direction, c, attrib, n_min, n_max, lower,
                          from_phrase):
    # TODO: this currently looks only in current table;
    #   precompute over the whole document/page instead
    bbox_direction_aligned = bbox_vert_aligned if direction == 'vert' else bbox_horz_aligned
    ngrams_space = Ngrams(n_max=n_max, split_tokens=[])
    f = (lambda w: w.lower()) if lower else (lambda w: w)
    spans = [c] if isinstance(c, TemporarySpan) else c.get_contexts()
    for span in spans:
        if not span.sentence.is_tabular() or not span.sentence.is_visual():
            continue
        for phrase in span.sentence.table.phrases:
            if (from_phrase):
                if (bbox_direction_aligned(
                        bbox_from_phrase(phrase), bbox_from_span(span))
                        and phrase is not span.sentence):
                    for ngram in tokens_to_ngrams(
                            getattr(phrase, attrib),
                            n_min=n_min,
                            n_max=n_max,
                            lower=lower):
                        yield ngram
            else:
                for ts in ngrams_space.apply(phrase):
                    if (bbox_direction_aligned(
                            bbox_from_span(ts), bbox_from_span(span))
                            and not (phrase == span.sentence
                                     and ts.get_span() in span.get_span())):
                        yield f(ts.get_span())


def get_vert_ngrams_left(c):
    """Not implemented."""
    # TODO
    return


def get_vert_ngrams_right(c):
    """Not implemented."""
    # TODO
    return


def get_vert_ngrams_center(c):
    """Not implemented."""
    # TODO
    return


def get_visual_header_ngrams(c, axis=None):
    """Not implemented."""
    # TODO
    return


def get_visual_distance(c, axis=None):
    """Not implemented."""
    # TODO
    return


# Default dimensions for 8.5" x 11"
DEFAULT_WIDTH = 612
DEFAULT_HEIGHT = 792


def get_page_vert_percentile(span,
                             page_width=DEFAULT_WIDTH,
                             page_height=DEFAULT_HEIGHT):
    """Return which percentile from the TOP in the page Span candidate is located in.

    Percentile is calculated where the top of the page is 0.0, and the bottom of
    the page is 1.0. For example, a Span in at the top 1/4 of the page will have
    a percentil of 0.25.

    Page width and height are based on pt values:
        Letter      612x792
        Tabloid     792x1224
        Ledger      1224x792
        Legal       612x1008
        Statement   396x612
        Executive   540x720
        A0          2384x3371
        A1          1685x2384
        A2          1190x1684
        A3          842x1190
        A4          595x842
        A4Small     595x842
        A5          420x595
        B4          729x1032
        B5          516x729
        Folio       612x936
        Quarto      610x780
        10x14       720x1008
    and should match the source documents. Letter size is used by default.

    Note that if a candidate is passed in, only the vertical percentil of its
    first Span is returned.

    :param span: The Span to evaluate
    :param page_width: The width of the page. Default to Letter paper width.
    :param page_height: The heigh of the page. Default to Letter paper height.
    :rtype: float in [0.0, 1.0]
    """
    span = span if isinstance(span, TemporarySpan) else span[0]
    return bbox_from_span(span).top / page_height


def get_page_horz_percentile(span,
                             page_width=DEFAULT_WIDTH,
                             page_height=DEFAULT_HEIGHT):
    """Return which percentile from the LEFT in the page the Span is located in.

    Percentile is calculated where the left of the page is 0.0, and the right of
    the page is 1.0.

    Page width and height are based on pt values:
        Letter      612x792
        Tabloid     792x1224
        Ledger      1224x792
        Legal       612x1008
        Statement   396x612
        Executive   540x720
        A0          2384x3371
        A1          1685x2384
        A2          1190x1684
        A3          842x1190
        A4          595x842
        A4Small     595x842
        A5          420x595
        B4          729x1032
        B5          516x729
        Folio       612x936
        Quarto      610x780
        10x14       720x1008
    and should match the source documents. Letter size is used by default.

    Note that if a candidate is passed in, only the vertical percentil of its
    first Span is returned.

    :param c: The Span to evaluate
    :param page_width: The width of the page. Default to Letter paper width.
    :param page_height: The heigh of the page. Default to Letter paper height.
    :rtype: float in [0.0, 1.0]
    """
    span = span if isinstance(span, TemporarySpan) else span[0]
    return bbox_from_span(span).left, page_width


def _assign_alignment_features(phrases_by_key, align_type):
    for key, phrases in phrases_by_key.items():
        if len(phrases) == 1:
            continue
        context_lemmas = set()
        for p in phrases:
            p._aligned_lemmas.update(context_lemmas)
            # update lemma context for upcoming phrases in the group
            if len(p.lemmas) < 7:
                new_lemmas = [
                    lemma.lower() for lemma in p.lemmas if lemma.isalpha()
                ]
                # if new_lemmas: print '++Lemmas for\t', p, context_lemmas
                context_lemmas.update(new_lemmas)
                context_lemmas.update(
                    align_type + lemma for lemma in new_lemmas)


def _preprocess_visual_features(doc):
    if hasattr(doc, '_visual_features'):
        return
    # cache flag
    doc._visual_features = True

    phrase_by_page = defaultdict(list)
    for phrase in doc.phrases:
        phrase_by_page[phrase.page[0]].append(phrase)
        phrase._aligned_lemmas = set()

    for page, phrases in phrase_by_page.items():
        # process per page alignments
        yc_aligned = defaultdict(list)
        x0_aligned = defaultdict(list)
        xc_aligned = defaultdict(list)
        x1_aligned = defaultdict(list)
        for phrase in phrases:
            phrase.bbox = bbox_from_phrase(phrase)
            phrase.yc = (phrase.bbox.top + phrase.bbox.bottom) / 2
            phrase.x0 = phrase.bbox.left
            phrase.x1 = phrase.bbox.right
            phrase.xc = (phrase.x0 + phrase.x1) / 2
            # index current phrase by different alignment keys
            yc_aligned[phrase.yc].append(phrase)
            x0_aligned[phrase.x0].append(phrase)
            x1_aligned[phrase.x1].append(phrase)
            xc_aligned[phrase.xc].append(phrase)
        for l in yc_aligned.values():
            l.sort(key=lambda p: p.xc)
        for l in x0_aligned.values():
            l.sort(key=lambda p: p.yc)
        for l in x1_aligned.values():
            l.sort(key=lambda p: p.yc)
        for l in xc_aligned.values():
            l.sort(key=lambda p: p.yc)
        _assign_alignment_features(yc_aligned, 'Y_')
        _assign_alignment_features(x0_aligned, 'LEFT_')
        _assign_alignment_features(x1_aligned, 'RIGHT_')
        _assign_alignment_features(xc_aligned, 'CENTER_')


def get_visual_aligned_lemmas(span):
    """Return a generator of the lemmas aligned visually with the Span.

    Note that if a candidate is passed in, all of its Spans will be searched.

    :param span: The Span to evaluate.
    :rtype: a _generator_ of lemmas
    """
    spans = [span] if isinstance(span, TemporarySpan) else span.get_contexts()
    for span in spans:
        phrase = span.sentence
        doc = phrase.document
        # cache features for the entire document
        _preprocess_visual_features(doc)

        for aligned_lemma in phrase._aligned_lemmas:
            yield aligned_lemma


def get_aligned_lemmas(span):
    """Return a set of the lemmas aligned visually with the Span.

    Note that if a candidate is passed in, all of its Spans will be searched.

    :param span: The Span to evaluate.
    :rtype: a set of lemmas
    """
    return set(get_visual_aligned_lemmas(span))


############################
# Structural feature helpers
############################
def get_tag(span):
    """Return the HTML tag of the Span.

    If a candidate is passed in, only the tag of its first Span is returned.

    These may be tags such as 'p', 'h2', 'table', 'div', etc.
    :param span: The Span to evaluate
    :rtype: string
    """
    span = span if isinstance(span, TemporarySpan) else span[0]
    return str(span.sentence.html_tag)


def get_attributes(span):
    """Return the HTML attributes of the Span.

    If a candidate is passed in, only the tag of its first Span is returned.

    A sample outout of this function on a Span in a paragraph tag is
    [u'style=padding-top: 8pt;padding-left: 20pt;text-indent: 0pt;text-align: left;']

    :param span: The Span to evaluate
    :rtype: list of strings representing HTML attributes
    """
    span = span if isinstance(span, TemporarySpan) else span[0]
    return span.sentence.html_attrs


# TODO: Too slow
def _get_node(phrase):
    return (etree.ElementTree(fromstring(phrase.document.text)).xpath(
        phrase.xpath))[0]


def get_parent_tag(span):
    """Return the HTML tag of the Span's parent.

    These may be tags such as 'p', 'h2', 'table', 'div', etc.
    If a candidate is passed in, only the tag of its first Span is returned.

    :param span: The Span to evaluate
    :rtype: string
    """
    span = span if isinstance(span, TemporarySpan) else span[0]
    i = _get_node(span.sentence)
    return str(i.getparent().tag) if i.getparent() is not None else None


def get_prev_sibling_tags(span):
    """Return the HTML tag of the Span's previous siblings.

    Previous siblings are Spans which are at the same level in the HTML tree as
    the given span, but are declared before the given span.
    If a candidate is passed in, only the previous siblings of its first Span
    are considered in the calculation.

    :param span: The Span to evaluate
    :rtype: list of strings
    """
    span = span if isinstance(span, TemporarySpan) else span[0]
    prev_sibling_tags = []
    i = _get_node(span.sentence)
    while i.getprevious() is not None:
        prev_sibling_tags.insert(0, str(i.getprevious().tag))
        i = i.getprevious()
    return prev_sibling_tags


def get_next_sibling_tags(span):
    """Return the HTML tag of the Span's next siblings.

    Next siblings are Spans which are at the same level in the HTML tree as
    the given span, but are declared after the given span.
    If a candidate is passed in, only the next siblings of its last Span
    are considered in the calculation.

    :param span: The Span to evaluate
    :rtype: list of strings
    """
    span = span if isinstance(span, TemporarySpan) else span[-1]
    next_sibling_tags = []
    i = _get_node(span.sentence)
    while i.getnext() is not None:
        next_sibling_tags.append(str(i.getnext().tag))
        i = i.getnext()
    return next_sibling_tags


def get_ancestor_class_names(span):
    """Return the HTML classes of the Span's ancestors.

    If a candidate is passed in, only the ancestors of its first Span are returned.

    :param span: The Span to evaluate
    :rtype: list of strings
    """
    span = span if isinstance(span, TemporarySpan) else span[0]
    class_names = []
    i = _get_node(span.sentence)
    while i is not None:
        class_names.insert(0, str(i.get('class')))
        i = i.getparent()
    return class_names


def get_ancestor_tag_names(span):
    """Return the HTML tag of the Span's ancestors.

    For example, ['html', 'body', 'p'].
    If a candidate is passed in, only the ancestors of its first Span are returned.

    :param span: The Span to evaluate
    :rtype: list of strings
    """
    span = span if isinstance(span, TemporarySpan) else span[0]
    tag_names = []
    i = _get_node(span.sentence)
    while i is not None:
        tag_names.insert(0, str(i.tag))
        i = i.getparent()
    return tag_names


def get_ancestor_id_names(span):
    """Return the HTML id's of the Span's ancestors.

    If a candidate is passed in, only the ancestors of its first Span are returned.

    :param span: The Span to evaluate
    :rtype: list of strings
    """
    span = span if isinstance(span, TemporarySpan) else span[0]
    id_names = []
    i = _get_node(span.sentence)
    while i is not None:
        id_names.insert(0, str(i.get('id')))
        i = i.getparent()
    return id_names


def common_ancestor(c):
    """Return the common path to the root that is shared between a binary-Span Candidate.

    In particular, this is the common path of HTML tags.

    :param c: The binary-Span Candidate to evaluate
    :rtype: list of strings
    """
    ancestor1 = np.array(c[0].sentence.xpath.split('/'))
    ancestor2 = np.array(c[1].sentence.xpath.split('/'))
    min_len = min(ancestor1.size, ancestor2.size)
    return list(
        ancestor1[:np.argmin(ancestor1[:min_len] == ancestor2[:min_len])])


def lowest_common_ancestor_depth(c):
    """Return the minimum distance between a binary-Span Candidate to their lowest common ancestor.

    For example, if the tree looked like this:

        html
        |----<div> span 1 </div>
        |----table
        |    |----tr
        |    |    |-----<th> span 2 </th>

    we return 1, the distance from span 1 to the html root. Smaller values indicate
    that two Spans are close structurally, while larger values indicate that two
    Spans are spread far apart structurally in the document.

    :param c: The binary-Span Candidate to evaluate
    :rtype: integer
    """
    ancestor1 = np.array(c[0].sentence.xpath.split('/'))
    ancestor2 = np.array(c[1].sentence.xpath.split('/'))
    min_len = min(ancestor1.size, ancestor2.size)
    return min_len - np.argmin(ancestor1[:min_len] == ancestor2[:min_len])



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import *

import numpy as np

from .rnn_base import RNNBase
from .utils import SymbolTable


class TextRNN(RNNBase):
    """TextRNN for strings of text."""
    def _preprocess_data(self, candidates, extend=False):
        """Convert candidate sentences to lookup sequences
        
        :param candidates: candidates to process
        :param extend: extend symbol table for tokens (train), or lookup (test)?
        """
        if not hasattr(self, 'word_dict'):
            self.word_dict = SymbolTable()
        data, ends = [], []
        for candidate in candidates:
            toks = candidate.get_contexts()[0].text.split()
            # Either extend word table or retrieve from it
            f = self.word_dict.get if extend else self.word_dict.lookup
            data.append(np.array(list(map(f, toks))))
            ends.append(len(toks))
        return data, ends



import os
import yaml


def dict2obj(d):
    if isinstance(d, dict):
        n = {}
        for item in d:
            if isinstance(d[item], dict):
                n[item] = dict2obj(d[item])
            elif isinstance(d[item], (list, tuple)):
                n[item] = [dict2obj(i) for i in d[item]]
            else:
                n[item] = d[item]
        return type('obj_from_dict', (object,), n)
    else:
        return d

# Load the settings relative to this config file.
directory = os.path.dirname(__file__)
with open(directory + "/" + "settings.yaml", "r") as f:
    settings = dict2obj(yaml.safe_load(f))



from __future__ import print_function
from __future__ import division
from builtins import str
from builtins import zip
from builtins import range
import codecs
import numpy as np
import scipy.sparse as sparse
import subprocess
import tempfile
from collections import namedtuple
from pandas import DataFrame, Series

from fonduer.snorkel.annotations import FeatureAnnotator
from fonduer.snorkel.models import Candidate
from fonduer.snorkel.models.meta import *
from fonduer.snorkel.udf import UDF, UDFRunner
from fonduer.snorkel.utils import (
    matrix_conflicts,
    matrix_coverage,
    matrix_overlaps,
    matrix_tp,
    matrix_fp,
    matrix_fn,
    matrix_tn
)
from fonduer.snorkel.utils import remove_files
from fonduer.features.features import get_all_feats

# Used to conform to existing annotation key API call
# Note that this anontation matrix class can not be replaced with snorkel one
# since we do not have ORM-backed key objects but rather a simple python list.
_TempKey = namedtuple('TempKey', ['id', 'name'])

def _to_annotation_generator(fns):
    """"
    Generic method which takes a set of functions, and returns a generator that yields
    function.__name__, function result pairs.
    """
    def fn_gen(c):
        for f in fns:
            yield f.__name__, f(c)
    return fn_gen

class csr_AnnotationMatrix(sparse.csr_matrix):
    """
    An extension of the scipy.sparse.csr_matrix class for holding sparse annotation matrices
    and related helper methods.
    """
    def __init__(self, arg1, **kwargs):
        # # Note: Currently these need to return None if unset, otherwise matrix copy operations break...
        # self.session = SnorkelSession()
        # Map candidate id to row id
        self.candidate_index = kwargs.pop('candidate_index', {})
        # Map row id to candidate id
        self.row_index = kwargs.pop('row_index', [])
        # Map col id to key str
        self.keys = kwargs.pop('keys', [])
        # Map key str to col number
        self.key_index = kwargs.pop('key_index', {})

        # Note that scipy relies on the first three letters of the class to define matrix type...
        super(csr_AnnotationMatrix, self).__init__(arg1, **kwargs)

    def get_candidate(self, session, i):
        """Return the Candidate object corresponding to row i"""
        return session.query(Candidate)\
                .filter(Candidate.id == self.row_index[i]).one()

    def get_row_index(self, candidate):
        """Return the row index of the Candidate"""
        return self.candidate_index[candidate.id]

    def get_key(self, j):
        """Return the AnnotationKey object corresponding to column j"""
        return _TempKey(j, self.keys[j])

    def get_col_index(self, key):
        """Return the cow index of the AnnotationKey"""
        return self.key_index[key.id]

    def stats(self):
        """Return summary stats about the annotations"""
        raise NotImplementedError()

    def lf_stats(self, labels=None, est_accs=None):
        """Returns a pandas DataFrame with the LFs and various per-LF statistics"""
        lf_names = self.keys

        # Default LF stats
        col_names = ['j', 'Coverage', 'Overlaps', 'Conflicts']
        d = {
            'j'         : list(range(self.shape[1])),
            'Coverage'  : Series(data=matrix_coverage(self), index=lf_names),
            'Overlaps'  : Series(data=matrix_overlaps(self), index=lf_names),
            'Conflicts' : Series(data=matrix_conflicts(self), index=lf_names)
        }
        if labels is not None:
            col_names.extend(['TP', 'FP', 'FN', 'TN', 'Empirical Acc.'])
            ls = np.ravel(labels.todense() if sparse.issparse(labels) else labels)
            tp = matrix_tp(self, ls)
            fp = matrix_fp(self, ls)
            fn = matrix_fn(self, ls)
            tn = matrix_tn(self, ls)
            ac = (tp+tn) / (tp+tn+fp+fn)
            d['Empirical Acc.'] = Series(data=ac, index=lf_names)
            d['TP']             = Series(data=tp, index=lf_names)
            d['FP']             = Series(data=fp, index=lf_names)
            d['FN']             = Series(data=fn, index=lf_names)
            d['TN']             = Series(data=tn, index=lf_names)

        if est_accs is not None:
            col_names.append('Learned Acc.')
            d['Learned Acc.'] = Series(data=est_accs, index=lf_names)
        return DataFrame(data=d, index=lf_names)[col_names]

segment_dir = tempfile.gettempdir()


def get_sql_name(text):
    """
    Create valid SQL identifier as part of a feature storage table name
    """
    # Normalize identifier
    text = ''.join(c.lower() if c.isalnum() else ' ' for c in text)
    text = '_'.join(text.split())
    return text


def tsv_escape(s):
    if s is None:
        return '\\N'
    # Make sure feature names are still uniquely encoded in ascii
    s = str(s)
    s = s.replace('\"', '\\\\"').replace('\t', '\\t')
    if any(c in ',{}' for c in s):
        s = '"' + s + '"'
    return s


def array_tsv_escape(vals):
    return '{' + ','.join(tsv_escape(p) for p in vals) + '}'


def table_exists(con, name):
    cur = con.execute("select exists(select * from information_schema.tables where table_name=%s)", (name,))
    return cur.fetchone()[0]


def copy_postgres(segment_file_blob, table_name, tsv_columns):
    """
    @var segment_file_blob: e.g. "segment_*.tsv"
    @var table_name: The SQL table name to copy into
    @var tsv_columns: a string listing column names in the segment files
    separated by comma. e.g. "name, age, income"
    """
    print('Copying %s to postgres' % table_name)
    if DBPORT:
        cmd = ('cat %s | psql -p %s %s -U %s -c "COPY %s(%s) '
                'FROM STDIN" --set=ON_ERROR_STOP=true') % \
                (segment_file_blob, DBPORT, DBNAME, DBUSER, table_name, tsv_columns)
    else:
        cmd = ('cat %s | psql %s -U %s -c "COPY %s(%s) '
                'FROM STDIN" --set=ON_ERROR_STOP=true') % \
                (segment_file_blob, DBNAME, DBUSER, table_name, tsv_columns)
    _out = subprocess.check_output(cmd, stderr=subprocess.STDOUT, shell=True)
    print(_out)


def _segment_filename(db_name, table_name, job_id, start = None, end = None):
    suffix = '*'
    if start is not None:
        suffix = str(start)
        if end is not None:
            suffix += '-' + str(end)
    return '%s_%s_%s_%s.tsv' % (db_name, table_name, job_id, suffix)


class COOFeatureAnnotator(FeatureAnnotator):
    def __init__(self, f=get_all_feats, **kwargs):
        super(COOFeatureAnnotator, f, **kwargs)


class BatchAnnotatorUDF(UDF):
    def __init__(self, f, **kwargs):
        self.anno_generator  = _to_annotation_generator(f) if hasattr(f, '__iter__') else f
        super(BatchAnnotatorUDF, self).__init__(**kwargs)

    def apply(self, batch_range, table_name, split, cache, **kwargs):
        """
        Applies a given function to a range of candidates

        Note: Accepts a id_range as argument, because of issues with putting Candidate subclasses
        into Queues (can't pickle...)
        """
        start, end = batch_range
        file_name = _segment_filename(DBNAME, table_name, split, self.worker_id)
        segment_path = os.path.join(segment_dir, file_name)
        candidates = self.session.query(Candidate).filter(Candidate.split == split).order_by(Candidate.id).slice(start, end)
        with codecs.open(segment_path, 'a+', encoding='utf-8') as writer:
            if not cache:
                for i, candidate in enumerate(candidates):
                    # Runs the actual extraction function
                    nonzero_kvs = [(k,v) for k, v in self.anno_generator(candidate) if v != 0]
                    if nonzero_kvs:
                        keys, values = list(zip(*nonzero_kvs))
                    else:
                        keys = values = []
                    row = [str(candidate.id), array_tsv_escape(keys), array_tsv_escape(values)]
                    writer.write('\t'.join(row) + '\n')
            else:
                nonzero_kv_dict = {}
                for id, k, v in self.anno_generator(list(candidates)):
                    if id not in nonzero_kv_dict: nonzero_kv_dict[id] = []
                    if v != 0: nonzero_kv_dict[id].append((k, v))
                for i, candidate in enumerate(candidates):
                    nonzero_kvs = nonzero_kv_dict[candidate.id]
                    if nonzero_kvs:
                        keys, values = list(zip(*nonzero_kvs))
                    else:
                        keys = values = []
                    row = [str(candidate.id), array_tsv_escape(keys), array_tsv_escape(values)]
                    writer.write('\t'.join(row) + '\n')
        return
        yield


class BatchAnnotator(UDFRunner):
    """Abstract class for annotating candidates and persisting these annotations to DB"""
    def __init__(self, candidate_type, annotation_type, f, batch_size = 50, **kwargs):
        if isinstance(candidate_type, type): candidate_type = candidate_type.__name__
        self.table_name = get_sql_name(candidate_type) + '_' + annotation_type
        self.key_table_name = self.table_name + '_keys'
        self.annotation_type = annotation_type
        self.batch_size = batch_size
        super(BatchAnnotator, self).__init__(BatchAnnotatorUDF, f=f, **kwargs)

    def apply(self, split, key_group=0, replace_key_set=True, update_keys=False, update_values=True, storage=None, ignore_keys=[], **kwargs):
        if update_keys: replace_key_set = False
        # Get the cids based on the split, and also the count
        SnorkelSession = new_sessionmaker()
        session   = SnorkelSession()
        # Note: In the current UDFRunner implementation, we load all these into memory and fill a
        # multiprocessing JoinableQueue with them before starting... so might as well load them here and pass in.
        # Also, if we try to pass in a query iterator instead, with AUTOCOMMIT on, we get a TXN error...
        candidates = session.query(Candidate).filter(Candidate.split == split).all()
        cids_count = len(candidates)
        if cids_count == 0:
            raise ValueError('No candidates in current split')

        # Setting up job batches
        chunks    = cids_count // self.batch_size
        batch_range = [(i * self.batch_size, (i + 1) * self.batch_size) for i in range(chunks)]
        remainder = cids_count % self.batch_size
        if remainder:
            batch_range.append((chunks * self.batch_size, cids_count))

        old_table_name = None
        table_name = self.table_name
        # Run the Annotator
        with snorkel_engine.connect() as con:
            table_already_exists = table_exists(con, table_name)
            if update_values and table_already_exists:
                # Now we extract under a temporary name for merging
                old_table_name = table_name
                table_name += '_updates'

            segment_file_blob = os.path.join(segment_dir, _segment_filename(DBNAME, self.table_name, split))
            remove_files(segment_file_blob)
            cache = True if self.annotation_type == 'feature' else False
            super(BatchAnnotator, self).apply(batch_range, table_name = self.table_name, split=split, cache=cache, **kwargs)

            # Insert and update keys
            if not table_already_exists or old_table_name:
                con.execute('CREATE TABLE %s(candidate_id integer PRIMARY KEY, keys text[] NOT NULL, values real[] NOT NULL)' % table_name)
            copy_postgres(segment_file_blob, table_name, 'candidate_id, keys, values')
            remove_files(segment_file_blob)

            # Replace the LIL table with COO if requested
            if storage == 'COO':
                temp_coo_table = table_name + '_COO'
                con.execute('CREATE TABLE %s AS '
                            '(SELECT candidate_id, UNNEST(keys) as key, UNNEST(values) as value from %s)' % (temp_coo_table, table_name))
                con.execute('DROP TABLE %s'%table_name)
                con.execute('ALTER TABLE %s RENAME TO %s' % (temp_coo_table, table_name))
                con.execute('ALTER TABLE %s ADD PRIMARY KEY(candidate_id, key)' % table_name)
                # Update old table
                if old_table_name:
                    con.execute('INSERT INTO %s SELECT * FROM %s ON CONFLICT(candidate_id, key) '
                                'DO UPDATE SET value=EXCLUDED.value'%(old_table_name, table_name))
                    con.execute('DROP TABLE %s' % table_name)
            else:# LIL
                # Update old table
                if old_table_name:
                    con.execute('INSERT INTO %s AS old SELECT * FROM %s ON CONFLICT(candidate_id) '
                                'DO UPDATE SET '
                                'values=old.values || EXCLUDED.values,'
                                'keys=old.keys || EXCLUDED.keys'%(old_table_name, table_name))
                    con.execute('DROP TABLE %s' % table_name)

            if old_table_name: table_name = old_table_name
            # Load the matrix
            key_table_name = self.key_table_name
            if key_group:
                key_table_name = self.key_table_name + '_' + get_sql_name(key_group)

            return load_annotation_matrix(con, candidates, split, table_name, key_table_name, replace_key_set, storage, update_keys, ignore_keys)

    def clear(self, session, split, replace_key_set = False, **kwargs):
        """
        Deletes the Annotations for the Candidates in the given split.
        If replace_key_set=True, deletes *all* Annotations (of this Annotation sub-class)
        and also deletes all AnnotationKeys (of this sub-class)
        """
        with snorkel_engine.connect() as con:
            if split is None:
                con.execute('DROP TABLE IF EXISTS %s' % self.table_name)
            elif table_exists(con, self.table_name):
                con.execute('DELETE FROM %s WHERE candidate_id IN '
                            '(SELECT id FROM candidate WHERE split=%d)' % (self.table_name, split))
            if replace_key_set:
                con.execute('DROP TABLE IF EXISTS %s' % self.key_table_name)

    def apply_existing(self, split, key_group=0, **kwargs):
        """Alias for apply that emphasizes we are using an existing AnnotatorKey set."""
        return self.apply(split, key_group=key_group, replace_key_set=False, **kwargs)

    def load_matrix(self, split, ignore_keys=[]):
        SnorkelSession = new_sessionmaker()
        session   = SnorkelSession()
        candidates = session.query(Candidate).filter(Candidate.split == split).all()
        with snorkel_engine.connect() as con:
            return load_annotation_matrix(con, candidates, split, self.table_name, self.key_table_name, False, None, False, ignore_keys)


class BatchFeatureAnnotator(BatchAnnotator):
    def __init__(self, candidate_type, **kwargs):
        super(BatchFeatureAnnotator, self).__init__(candidate_type, annotation_type='feature', f=get_all_feats, **kwargs)


class BatchLabelAnnotator(BatchAnnotator):
    def __init__(self, candidate_type, lfs, **kwargs):
        if lfs is not None:
            labels = lambda c : [(lf.__name__, lf(c)) for lf in lfs]
        elif label_generator is not None:
            labels = lambda c : label_generator(c)
        else:
            raise ValueError("Must provide lfs or label_generator kwarg.")

        # Convert lfs to a generator function
        # In particular, catch verbose values and convert to integer ones
        def f_gen(c):
            for lf_key, label in labels(c):
                # Note: We assume if the LF output is an int, it is already
                # mapped correctly
                if isinstance(label, int):
                    yield lf_key, label
                # None is a protected LF output value corresponding to 0,
                # representing LF abstaining
                elif label is None:
                    yield lf_key, 0
                elif label in c.values:
                    if c.cardinality > 2:
                        yield lf_key, c.values.index(label) + 1
                    # Note: Would be nice to not special-case here, but for
                    # consistency we leave binary LF range as {-1,0,1}
                    else:
                        val = 1 if c.values.index(label) == 0 else -1
                        yield lf_key, val
                else:
                    raise ValueError("""
                        Unable to parse label with value %s
                        for candidate with values %s""" % (label, c.values))
        super(BatchLabelAnnotator, self).__init__(candidate_type, annotation_type='label', f=f_gen, **kwargs)


def load_annotation_matrix(con, candidates, split, table_name, key_table_name, replace_key_set, storage, update_keys, ignore_keys):
    """
    Loads a sparse matrix from an annotation table
    """
    if replace_key_set:
        # Recalculate unique keys for this set of candidates
        con.execute('DROP TABLE IF EXISTS %s' % key_table_name)
    if replace_key_set or not table_exists(con, key_table_name):
        if storage == 'COO':
            con.execute('CREATE TABLE %s AS '
                        '(SELECT DISTINCT key FROM %s)' % (key_table_name, table_name))
        else:
            con.execute('CREATE TABLE %s AS '
                        '(SELECT DISTINCT UNNEST(keys) as key FROM %s)' % (key_table_name, table_name))
        con.execute('ALTER TABLE %s ADD PRIMARY KEY(key)' % key_table_name)
    elif update_keys:
        if storage == 'COO':
            con.execute('INSERT INTO %s SELECT DISTINCT key FROM %s '
                        'ON CONFLICT(key) DO NOTHING' % (key_table_name, table_name))
        else:
            con.execute('INSERT INTO %s SELECT DISTINCT UNNEST(keys) as key FROM %s '
                        'ON CONFLICT(key) DO NOTHING' % (key_table_name, table_name))

    # The result should be a list of all feature strings, small enough to hold in memory
    # TODO: store the actual index in table in case row number is unstable between queries
    ignore_keys = set(ignore_keys)
    keys = [row[0] for row in con.execute('SELECT * FROM %s' % key_table_name) if row[0] not in ignore_keys]
    key_index = {key:i for i, key in enumerate(keys)}
    # Create sparse matrix in LIL format for incremental construction
    lil_feat_matrix = sparse.lil_matrix((len(candidates), len(keys)), dtype=np.int64)

    row_index = []
    candidate_index = {}
    # Load annotations from database
    # TODO: move this for-loop computation to database for automatic parallelization,
    # avoid communication overhead etc. Try to avoid the log sorting factor using unnest
    if storage == 'COO':
        print('key size', len(keys))
        print('candidate size', len(candidates))
        iterator_sql = 'SELECT candidate_id, key, value FROM %s '
        'WHERE candidate_id IN '
        '(SELECT id FROM candidate WHERE split=%d) '
        'ORDER BY candidate_id' % (table_name, split)
        prev_id = None
        i = -1
        for _, (candidate_id, key, value) in enumerate(con.execute(iterator_sql)):
            # Update candidate index tracker
            if candidate_id != prev_id:
                i += 1
                candidate_index[candidate_id] = i
                row_index.append(candidate_id)
                prev_id = candidate_id
            # Only keep known features
            key_id = key_index.get(key, None)
            if key_id is not None:
                lil_feat_matrix[i, key_id] = int(value)

    else:
        iterator_sql = '''SELECT candidate_id, keys, values FROM %s
                          WHERE candidate_id IN
                          (SELECT id FROM candidate WHERE split=%d)
                          ORDER BY candidate_id''' % (table_name, split)
        for i, (candidate_id, c_keys, values) in enumerate(con.execute(iterator_sql)):
            candidate_index[candidate_id] = i
            row_index.append(candidate_id)
            for key, value in zip(c_keys, values):
                # Only keep known features
                key_id = key_index.get(key, None)
                if key_id is not None:
                    lil_feat_matrix[i, key_id] = int(value)

    return csr_AnnotationMatrix(lil_feat_matrix, candidate_index=candidate_index,
                                    row_index=row_index, keys=keys, key_index=key_index)
